@@ -1,3 +1,10 @@
+{-# OPTIONS -w #-}
+-- The above warning supression flag is a temporary kludge.
+-- While working on this module you are encouraged to remove it and fix
+-- any warnings in the module. See
+--     http://hackage.haskell.org/trac/ghc/wiki/Commentary/CodingStyle#Warnings
+-- for details
+
 -----------------------------------------------------------------------------
 --  $Id: DriverPhases.hs,v 1.38 2005/05/17 11:01:59 simonmar Exp $
 --
@@ -183,10 +190,6 @@ phaseInputExt SplitAs     	  = "split_s"   -- not really generated
 phaseInputExt CmmCpp	  	  = "cmm"
 phaseInputExt Cmm	  	  = "cmmcpp"
 phaseInputExt StopLn          	  = "o"
-#ifdef ILX
-phaseInputExt Ilx2Il      	  = "ilx"
-phaseInputExt Ilasm       	  = "il"
-#endif
 
 haskellish_src_suffixes      = haskellish_user_src_suffixes ++
 			       [ "hspp", "hscpp", "hcr", "cmm" ]
