@@ -1,3 +1,10 @@
+{-# OPTIONS -w #-}
+-- The above warning supression flag is a temporary kludge.
+-- While working on this module you are encouraged to remove it and fix
+-- any warnings in the module. See
+--     http://hackage.haskell.org/trac/ghc/wiki/Commentary/CodingStyle#Warnings
+-- for details
+
 -----------------------------------------------------------------------------
 --
 -- GHC Driver
@@ -12,16 +19,13 @@ module DriverPipeline (
    oneShot, compileFile,
 
 	-- Interfaces for the batch-mode driver
-   staticLink,
+   linkBinary,
 
 	-- Interfaces for the compilation manager (interpreted/batch-mode)
    preprocess, 
    compile, CompResult(..), 
    link, 
 
-        -- DLL building
-   doMkDLL,
-
   ) where
 
 #include "HsVersions.h"
@@ -30,7 +34,6 @@ import Packages
 import HeaderInfo
 import DriverPhases
 import SysTools
-import qualified SysTools	
 import HscMain
 import Finder
 import HscTypes
@@ -39,7 +42,7 @@ import Module
 import UniqFM		( eltsUFM )
 import ErrUtils
 import DynFlags
-import StaticFlags	( v_Ld_inputs, opt_Static, WayName(..) )
+import StaticFlags	( v_Ld_inputs, opt_Static, opt_HardwireLibPaths, WayName(..) )
 import Config
 import Panic
 import Util
@@ -50,17 +53,17 @@ import ParserCoreUtils	( getCoreModuleName )
 import SrcLoc		( unLoc )
 import SrcLoc		( Located(..) )
 
-import EXCEPTION
-import DATA_IOREF	( readIORef, writeIORef, IORef )
-import GLAEXTS		( Int(..) )
-
-import Directory
-import System
-import IO
-import Monad
+import Control.Exception as Exception
+import Data.IORef	( readIORef, writeIORef, IORef )
+import GHC.Exts		( Int(..) )
+import System.Directory
+import System.IO
+import SYSTEM_IO_ERROR as IO
+import Control.Monad
 import Data.List	( isSuffixOf )
-import Maybe
-
+import Data.Maybe
+import System.Exit
+import System.Environment
 
 -- ---------------------------------------------------------------------------
 -- Pre-process
@@ -160,7 +163,10 @@ compile hsc_env mod_summary maybe_old_linkable old_iface mod_index nmods = do
              return (CompOK details iface maybe_old_linkable)
        handleBatch (HscRecomp hasStub, iface, details)
            | isHsBoot src_flavour
-               = return (CompOK details iface Nothing)
+               = do when (isObjectTarget hsc_lang) $ -- interpreted reaches here too
+                       SysTools.touch dflags' "Touching object file"
+                                   object_filename
+                    return (CompOK details iface Nothing)
            | otherwise
                = do stub_unlinked <- getStubLinkable hasStub
                     (hs_unlinked, unlinked_time) <-
@@ -183,9 +189,9 @@ compile hsc_env mod_summary maybe_old_linkable old_iface mod_index nmods = do
        handleInterpreted (InteractiveNoRecomp, iface, details)
            = ASSERT (isJust maybe_old_linkable)
              return (CompOK details iface maybe_old_linkable)
-       handleInterpreted (InteractiveRecomp hasStub comp_bc, iface, details)
+       handleInterpreted (InteractiveRecomp hasStub comp_bc modBreaks, iface, details)
            = do stub_unlinked <- getStubLinkable hasStub
-                let hs_unlinked = [BCOs comp_bc]
+                let hs_unlinked = [BCOs comp_bc modBreaks]
                     unlinked_time = ms_hs_date mod_summary
                   -- Why do we use the timestamp of the source file here,
                   -- rather than the current time?  This works better in
@@ -206,13 +212,11 @@ compile hsc_env mod_summary maybe_old_linkable old_iface mod_index nmods = do
                   Just result -> handle result
    -- run the compiler
    case hsc_lang of
-     HscInterpreted | not (isHsBoot src_flavour) -- We can't compile boot files to
-                                                 -- bytecode so don't even try.
-         -> runCompiler hscCompileInteractive handleInterpreted
-     HscNothing
-         -> runCompiler hscCompileNothing handleBatch
-     _other
-         -> runCompiler hscCompileBatch handleBatch
+     HscInterpreted
+      | isHsBoot src_flavour -> runCompiler hscCompileNothing handleBatch
+      | otherwise            -> runCompiler hscCompileInteractive handleInterpreted
+     HscNothing     -> runCompiler hscCompileNothing handleBatch
+     _other         -> runCompiler hscCompileBatch handleBatch
 
 -----------------------------------------------------------------------------
 -- stub .h and .c files (for foreign export support)
@@ -239,7 +243,7 @@ compileStub dflags mod location = do
 	    stub_o = o_base ++ "_stub" `joinFileExt` o_ext
 
 	-- compile the _stub.c file w/ gcc
-	let (stub_c,_) = mkStubPaths dflags (moduleName mod) location
+	let (stub_c,_,_) = mkStubPaths dflags (moduleName mod) location
 	runPipeline StopLn dflags (stub_c,Nothing)  Nothing
 		(SpecificFile stub_o) Nothing{-no ModLocation-}
 
@@ -249,10 +253,10 @@ compileStub dflags mod location = do
 -- ---------------------------------------------------------------------------
 -- Link
 
-link :: GhcMode			-- interactive or batch
-     -> DynFlags		-- dynamic flags
-     -> Bool			-- attempt linking in batch mode?
-     -> HomePackageTable	-- what to link
+link :: GhcLink                 -- interactive or batch
+     -> DynFlags                -- dynamic flags
+     -> Bool                    -- attempt linking in batch mode?
+     -> HomePackageTable        -- what to link
      -> IO SuccessFlag
 
 -- For the moment, in the batch linker, we don't bother to tell doLink
@@ -263,72 +267,74 @@ link :: GhcMode			-- interactive or batch
 -- will succeed.
 
 #ifdef GHCI
-link Interactive dflags batch_attempt_linking hpt
+link LinkInMemory dflags batch_attempt_linking hpt
     = do -- Not Linking...(demand linker will do the job)
-	 return Succeeded
+         return Succeeded
 #endif
 
-link JustTypecheck dflags batch_attempt_linking hpt
+link NoLink dflags batch_attempt_linking hpt
    = return Succeeded
 
-link BatchCompile dflags batch_attempt_linking hpt
+link LinkBinary dflags batch_attempt_linking hpt
    | batch_attempt_linking
-   = do 
-	let 
-	    home_mod_infos = eltsUFM hpt
+   = do
+        let
+            home_mod_infos = eltsUFM hpt
 
-	    -- the packages we depend on
-	    pkg_deps  = concatMap (dep_pkgs . mi_deps . hm_iface) home_mod_infos
+            -- the packages we depend on
+            pkg_deps  = concatMap (dep_pkgs . mi_deps . hm_iface) home_mod_infos
 
-	    -- the linkables to link
-	    linkables = map (expectJust "link".hm_linkable) home_mod_infos
+            -- the linkables to link
+            linkables = map (expectJust "link".hm_linkable) home_mod_infos
 
         debugTraceMsg dflags 3 (text "link: linkables are ..." $$ vcat (map ppr linkables))
 
-	-- check for the -no-link flag
-	if isNoLink (ghcLink dflags)
-	  then do debugTraceMsg dflags 3 (text "link(batch): linking omitted (-c flag given).")
-	          return Succeeded
-	  else do
+        -- check for the -no-link flag
+        if isNoLink (ghcLink dflags)
+          then do debugTraceMsg dflags 3 (text "link(batch): linking omitted (-c flag given).")
+                  return Succeeded
+          else do
 
-	let getOfiles (LM _ _ us) = map nameOfObject (filter isObject us)
-	    obj_files = concatMap getOfiles linkables
+        let getOfiles (LM _ _ us) = map nameOfObject (filter isObject us)
+            obj_files = concatMap getOfiles linkables
 
-	    exe_file = exeFileName dflags
+            exe_file = exeFileName dflags
 
-	-- if the modification time on the executable is later than the
-	-- modification times on all of the objects, then omit linking
-	-- (unless the -no-recomp flag was given).
-	e_exe_time <- IO.try $ getModificationTime exe_file
-	let linking_needed 
-		| Left _  <- e_exe_time = True
-		| Right t <- e_exe_time = 
-			any (t <) (map linkableTime linkables)
+        -- if the modification time on the executable is later than the
+        -- modification times on all of the objects, then omit linking
+        -- (unless the -no-recomp flag was given).
+        e_exe_time <- IO.try $ getModificationTime exe_file
+        extra_ld_inputs <- readIORef v_Ld_inputs
+        extra_times <- mapM (IO.try . getModificationTime) extra_ld_inputs
+        let other_times = map linkableTime linkables
+                       ++ [ t' | Right t' <- extra_times ]
+            linking_needed
+                | Left _  <- e_exe_time = True
+                | Right t <- e_exe_time = any (t <) other_times
 
-	if not (dopt Opt_ForceRecomp dflags) && not linking_needed
-	   then do debugTraceMsg dflags 2 (text exe_file <+> ptext SLIT("is up to date, linking not required."))
-		   return Succeeded
-	   else do
+        if not (dopt Opt_ForceRecomp dflags) && not linking_needed
+           then do debugTraceMsg dflags 2 (text exe_file <+> ptext SLIT("is up to date, linking not required."))
+                   return Succeeded
+           else do
 
-	debugTraceMsg dflags 1 (ptext SLIT("Linking") <+> text exe_file
-				 <+> text "...")
+        debugTraceMsg dflags 1 (ptext SLIT("Linking") <+> text exe_file
+                                 <+> text "...")
 
-	-- Don't showPass in Batch mode; doLink will do that for us.
-	let link = case ghcLink dflags of
-	        MkDLL       -> doMkDLL
-	        StaticLink  -> staticLink
-	link dflags obj_files pkg_deps
+        -- Don't showPass in Batch mode; doLink will do that for us.
+        let link = case ghcLink dflags of
+                LinkBinary  -> linkBinary
+                LinkDynLib  -> linkDynLib
+        link dflags obj_files pkg_deps
 
         debugTraceMsg dflags 3 (text "link: done")
 
-	-- staticLink only returns if it succeeds
+        -- linkBinary only returns if it succeeds
         return Succeeded
 
    | otherwise
    = do debugTraceMsg dflags 3 (text "link(batch): upsweep (partially) failed OR" $$
                                 text "   Main.main not exported; not linking.")
         return Succeeded
-      
 
 -- -----------------------------------------------------------------------------
 -- Compile files in one-shot mode.
@@ -376,8 +382,8 @@ doLink dflags stop_phase o_files
   | otherwise
   = case ghcLink dflags of
 	NoLink     -> return ()
-	StaticLink -> staticLink dflags o_files link_pkgs
-	MkDLL      -> doMkDLL dflags o_files link_pkgs
+	LinkBinary -> linkBinary dflags o_files link_pkgs
+	LinkDynLib -> linkDynLib dflags o_files []
   where
    -- Always link in the haskell98 package for static linking.  Other
    -- packages have to be specified via the -package flag.
@@ -415,16 +421,18 @@ runPipeline
   -> Maybe ModLocation          -- A ModLocation, if this is a Haskell module
   -> IO (DynFlags, FilePath)	-- (final flags, output filename)
 
-runPipeline stop_phase dflags (input_fn, mb_phase) mb_basename output maybe_loc
+runPipeline stop_phase dflags0 (input_fn, mb_phase) mb_basename output maybe_loc
   = do
-  let (input_basename, suffix) = splitFilename input_fn
+  let
+      (input_basename, suffix) = splitFilename input_fn
       basename | Just b <- mb_basename = b
                | otherwise             = input_basename
 
+      -- Decide where dump files should go based on the pipeline output
+      dflags = dflags0 { dumpPrefix = Just (basename ++ ".") }
+
 	-- If we were given a -x flag, then use that phase to start from
-      start_phase
-	| Just x_phase <- mb_phase = x_phase
-	| otherwise                = startPhase suffix
+      start_phase = fromMaybe (startPhase suffix) mb_phase
 
   -- We want to catch cases of "you can't get there from here" before
   -- we start the pipeline, because otherwise it will just run off the
@@ -636,7 +644,7 @@ runPhase (Hsc src_flavour) stop dflags0 basename suff input_fn get_output_fn _ma
  = do	-- normal Hsc mode, not mkdependHS
 
   -- we add the current directory (i.e. the directory in which
-  -- the .hs files resides) to the import path, since this is
+  -- the .hs files resides) to the include path, since this is
   -- what gcc does, and it's probably what you want.
 	let current_dir = directoryOf basename
 	
@@ -644,15 +652,15 @@ runPhase (Hsc src_flavour) stop dflags0 basename suff input_fn get_output_fn _ma
 	    dflags = dflags0 { includePaths = current_dir : paths }
 	
   -- gather the imports and module name
-        (hspp_buf,mod_name) <- 
+        (hspp_buf,mod_name,imps,src_imps) <- 
             case src_flavour of
 		ExtCoreFile -> do {  -- no explicit imports in ExtCore input.
 			          ; m <- getCoreModuleName input_fn
-			          ; return (Nothing, mkModuleName m) }
+			          ; return (Nothing, mkModuleName m, [], []) }
 
 		other -> do { buf <- hGetStringBuffer input_fn
-			    ; (_,_,L _ mod_name) <- getImports dflags buf input_fn
-			    ; return (Just buf, mod_name) }
+			    ; (src_imps,imps,L _ mod_name) <- getImports dflags buf input_fn (basename `joinFileExt` suff)
+			    ; return (Just buf, mod_name, imps, src_imps) }
 
   -- Build a ModLocation to pass to hscMain.
   -- The source filename is rather irrelevant by now, but it's used
@@ -739,8 +747,8 @@ runPhase (Hsc src_flavour) stop dflags0 basename suff input_fn get_output_fn _ma
 					ms_location  = location4,
 					ms_hs_date   = src_timestamp,
 					ms_obj_date  = Nothing,
-					ms_imps	     = unused_field,
-					ms_srcimps   = unused_field }
+					ms_imps	     = imps,
+					ms_srcimps   = src_imps }
 
   -- run the compiler!
 	mbResult <- hscCompileOneShot hsc_env
@@ -815,6 +823,7 @@ runPhase cc_phase stop dflags basename suff input_fn get_output_fn maybe_loc
 			      (cmdline_include_paths ++ pkg_include_dirs)
 
 	let (md_c_flags, md_regd_c_flags) = machdepCCOpts dflags
+        gcc_extra_viac_flags <- getExtraViaCOpts dflags
         let pic_c_flags = picCCOpts dflags
 
         let verb = getVerbFlag dflags
@@ -881,6 +890,13 @@ runPhase cc_phase stop dflags basename suff input_fn get_output_fn maybe_loc
 		       ++ (if hcc && mangle
 		  	     then md_regd_c_flags
 		  	     else [])
+		       ++ (if hcc
+		  	     then if mangle 
+                                     then gcc_extra_viac_flags
+                                     else filter (=="-fwrapv")
+                                                gcc_extra_viac_flags
+                                -- still want -fwrapv even for unreg'd
+		  	     else [])
 		       ++ (if hcc 
 			     then more_hcc_opts
 			     else [])
@@ -890,10 +906,6 @@ runPhase cc_phase stop dflags basename suff input_fn get_output_fn maybe_loc
 		       ++ split_opt
 		       ++ include_paths
 		       ++ pkg_extra_cc_opts
-#ifdef HAVE_GCC_HAS_WRAPV
-                  -- We need consistent integer overflow (trac #952)
-               ++ ["-fwrapv"]
-#endif
 		       ))
 
 	return (next_phase, dflags, maybe_loc, output_fn)
@@ -1053,7 +1065,7 @@ runPhase SplitAs stop dflags basename _suff _input_fn get_output_fn maybe_loc
 -- wrapper script calling the binary. Currently, we need this only in 
 -- a parallel way (i.e. in GUM), because PVM expects the binary in a
 -- central directory.
--- This is called from staticLink below, after linking. I haven't made it
+-- This is called from linkBinary below, after linking. I haven't made it
 -- a separate phase to minimise interfering with other modules, and
 -- we don't need the generality of a phase (MoveBinary is always
 -- done after linking and makes only sense in a parallel setup)   -- HWL
@@ -1067,9 +1079,9 @@ runPhase_MoveBinary dflags input_fn
            pvm_executable_base = "=" ++ input_fn
            pvm_executable = pvm_root ++ "/bin/" ++ pvm_arch ++ "/" ++ pvm_executable_base
         -- nuke old binary; maybe use configur'ed names for cp and rm?
-        system ("rm -f " ++ pvm_executable)
+        Panic.try (removeFile pvm_executable)
         -- move the newly created binary into PVM land
-        system ("cp -p " ++ input_fn ++ " " ++ pvm_executable)
+        copy dflags "copying PVM executable" input_fn pvm_executable
         -- generate a wrapper script for running a parallel prg under PVM
         writeFile input_fn (mk_pvm_wrapper_script pvm_executable pvm_executable_base sysMan)
 	return True
@@ -1107,18 +1119,18 @@ mk_pvm_wrapper_script pvm_executable pvm_executable_base sysMan = unlines $
   "",
   "args: while ($a = shift(@ARGV)) {",
   "    if ( $a eq '+RTS' ) {",
-  "	$in_RTS_args = 1;",
+  "        $in_RTS_args = 1;",
   "    } elsif ( $a eq '-RTS' ) {",
-  "	$in_RTS_args = 0;",
+  "        $in_RTS_args = 0;",
   "    }",
   "    if ( $a eq '-d' && $in_RTS_args ) {",
-  "	$debug = '-';",
+  "        $debug = '-';",
   "    } elsif ( $a =~ /^-qN(\\d+)/ && $in_RTS_args ) {",
-  "	$nprocessors = $1;",
+  "        $nprocessors = $1;",
   "    } elsif ( $a =~ /^-qp(\\d+)/ && $in_RTS_args ) {",
-  "	$nprocessors = $1;",
+  "        $nprocessors = $1;",
   "    } else {",
-  "	push(@nonPVM_args, $a);",
+  "        push(@nonPVM_args, $a);",
   "    }",
   "}",
   "",
@@ -1146,7 +1158,7 @@ checkProcessArgsResult flags filename
 
 getHCFilePackages :: FilePath -> IO [PackageId]
 getHCFilePackages filename =
-  EXCEPTION.bracket (openFile filename ReadMode) hClose $ \h -> do
+  Exception.bracket (openFile filename ReadMode) hClose $ \h -> do
     l <- hGetLine h
     case l of
       '/':'*':' ':'G':'H':'C':'_':'P':'A':'C':'K':'A':'G':'E':'S':rest ->
@@ -1167,8 +1179,8 @@ getHCFilePackages filename =
 -- read any interface files), so the user must explicitly specify all
 -- the packages.
 
-staticLink :: DynFlags -> [FilePath] -> [PackageId] -> IO ()
-staticLink dflags o_files dep_packages = do
+linkBinary :: DynFlags -> [FilePath] -> [PackageId] -> IO ()
+linkBinary dflags o_files dep_packages = do
     let verb = getVerbFlag dflags
         output_fn = exeFileName dflags
 
@@ -1177,7 +1189,9 @@ staticLink dflags o_files dep_packages = do
     -- dependencies, and eliminating duplicates.
 
     pkg_lib_paths <- getPackageLibraryPath dflags dep_packages
-    let pkg_lib_path_opts = map ("-L"++) pkg_lib_paths
+    let pkg_lib_path_opts = concat (map get_pkg_lib_path_opts pkg_lib_paths)
+	get_pkg_lib_path_opts l | opt_HardwireLibPaths && not opt_Static = ["-L" ++ l, "-Wl,-rpath", "-Wl," ++ l]
+				| otherwise = ["-L" ++ l]
 
     let lib_paths = libraryPaths dflags
     let lib_path_opts = map ("-L"++) lib_paths
@@ -1229,6 +1243,8 @@ staticLink dflags o_files dep_packages = do
 			]
 		    | otherwise               = []
 
+    rc_objs <- maybeCreateManifest dflags output_fn
+
     let (md_c_flags, _) = machdepCCOpts dflags
     SysTools.runLink dflags ( 
 		       [ SysTools.Option verb
@@ -1241,6 +1257,7 @@ staticLink dflags o_files dep_packages = do
 		      ++ extra_ld_inputs
 	 	      ++ lib_path_opts
 	 	      ++ extra_ld_opts
+                      ++ rc_objs
 #ifdef darwin_TARGET_OS
 	 	      ++ framework_path_opts
 	 	      ++ framework_opts
@@ -1279,16 +1296,67 @@ exeFileName dflags
 	"a.out"
 #endif
 
------------------------------------------------------------------------------
--- Making a DLL (only for Win32)
+maybeCreateManifest
+   :: DynFlags
+   -> FilePath                          -- filename of executable
+   -> IO [FilePath]                     -- extra objects to embed, maybe
+maybeCreateManifest dflags exe_filename = do
+#ifndef mingw32_TARGET_OS
+  return []
+#else
+  if not (dopt Opt_GenManifest dflags) then return [] else do
 
-doMkDLL :: DynFlags -> [String] -> [PackageId] -> IO ()
-doMkDLL dflags o_files dep_packages = do
+  let manifest_filename = exe_filename `joinFileExt` "manifest"
+
+  writeFile manifest_filename $ 
+      "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"yes\"?>\n"++
+      "  <assembly xmlns=\"urn:schemas-microsoft-com:asm.v1\" manifestVersion=\"1.0\">\n"++
+      "  <assemblyIdentity version=\"1.0.0.0\"\n"++
+      "     processorArchitecture=\"X86\"\n"++
+      "     name=\"" ++ basenameOf exe_filename ++ "\"\n"++
+      "     type=\"win32\"/>\n\n"++
+      "  <trustInfo xmlns=\"urn:schemas-microsoft-com:asm.v3\">\n"++
+      "    <security>\n"++
+      "      <requestedPrivileges>\n"++
+      "        <requestedExecutionLevel level=\"asInvoker\" uiAccess=\"false\"/>\n"++
+      "        </requestedPrivileges>\n"++
+      "       </security>\n"++
+      "  </trustInfo>\n"++
+      "</assembly>\n"
+
+  -- Windows will fine the manifest file if it is named foo.exe.manifest.
+  -- However, for extra robustness, and so that we can move the binary around,
+  -- we can embed the manifest in the binary itself using windres:
+  if not (dopt Opt_EmbedManifest dflags) then return [] else do
+
+  rc_filename <- newTempName dflags "rc"
+  rc_obj_filename <- newTempName dflags (objectSuf dflags)
+
+  writeFile rc_filename $
+      "1 24 MOVEABLE PURE " ++ show manifest_filename ++ "\n"
+        -- magic numbers :-)
+        -- show is a bit hackish above, but we need to esacpe the
+        -- backslashes in the path.
+
+  let wr_opts = getOpts dflags opt_windres
+  runWindres dflags $ map SysTools.Option $
+        ["--input="++rc_filename, 
+         "--output="++rc_obj_filename,
+         "--output-format=coff"] 
+        ++ wr_opts
+        -- no FileOptions here: windres doesn't like seeing
+        -- backslashes, apparently
+
+  return [rc_obj_filename]
+#endif
+
+
+linkDynLib :: DynFlags -> [String] -> [PackageId] -> IO ()
+linkDynLib dflags o_files dep_packages = do
     let verb = getVerbFlag dflags
     let static = opt_Static
     let no_hs_main = dopt Opt_NoHsMain dflags
     let o_file = outputFile dflags
-    let output_fn = case o_file of { Just s -> s; Nothing -> "HSdll.dll"; }
 
     pkg_lib_paths <- getPackageLibraryPath dflags dep_packages
     let pkg_lib_path_opts = map ("-L"++) pkg_lib_paths
@@ -1301,20 +1369,88 @@ doMkDLL dflags o_files dep_packages = do
 	-- probably _stub.o files
     extra_ld_inputs <- readIORef v_Ld_inputs
 
-	-- opts from -optdll-<blah>
-    let extra_ld_opts = getOpts dflags opt_dll 
-
-    let pstate = pkgState dflags
-	rts_pkg  = getPackageDetails pstate rtsPackageId
-        base_pkg = getPackageDetails pstate basePackageId
-
-    let extra_os = if static || no_hs_main
-                   then []
-                   else [ head (libraryDirs rts_pkg) ++ "/Main.dll_o",
-                          head (libraryDirs base_pkg) ++ "/PrelMain.dll_o" ]
-
     let (md_c_flags, _) = machdepCCOpts dflags
-    SysTools.runMkDLL dflags
+    let extra_ld_opts = getOpts dflags opt_l
+#if defined(mingw32_HOST_OS)
+    -----------------------------------------------------------------------------
+    -- Making a DLL
+    -----------------------------------------------------------------------------
+    let output_fn = case o_file of { Just s -> s; Nothing -> "HSdll.dll"; }
+
+    SysTools.runLink dflags
+	 ([ SysTools.Option verb
+	  , SysTools.Option "-o"
+	  , SysTools.FileOption "" output_fn
+	  , SysTools.Option "-shared"
+	  , SysTools.FileOption "-Wl,--out-implib=" (output_fn ++ ".a")
+	  ]
+	 ++ map (SysTools.FileOption "") o_files
+	 ++ map SysTools.Option (
+	    md_c_flags
+	 ++ extra_ld_inputs
+	 ++ lib_path_opts
+	 ++ extra_ld_opts
+	 ++ pkg_lib_path_opts
+	 ++ pkg_link_opts
+	))
+#elif defined(darwin_TARGET_OS)
+    -----------------------------------------------------------------------------
+    -- Making a darwin dylib
+    -----------------------------------------------------------------------------
+    -- About the options used for Darwin:
+    -- -dynamiclib
+    --   Apple's way of saying -shared
+    -- -undefined dynamic_lookup:
+    --   Without these options, we'd have to specify the correct dependencies
+    --   for each of the dylibs. Note that we could (and should) do without this
+    --	 for all libraries except the RTS; all we need to do is to pass the
+    --	 correct HSfoo_dyn.dylib files to the link command.
+    --	 This feature requires Mac OS X 10.3 or later; there is a similar feature,
+    --	 -flat_namespace -undefined suppress, which works on earlier versions,
+    --	 but it has other disadvantages.
+    -- -single_module
+    --	 Build the dynamic library as a single "module", i.e. no dynamic binding
+    --	 nonsense when referring to symbols from within the library. The NCG
+    --	 assumes that this option is specified (on i386, at least).
+    -- -Wl,-macosx_version_min -Wl,10.3
+    --	 Tell the linker its safe to assume that the library will run on 10.3 or
+    --	 later, so that it will not complain about the use of the option
+    --	 -undefined dynamic_lookup above.
+    -- -install_name
+    --   Causes the dynamic linker to ignore the DYLD_LIBRARY_PATH when loading
+    --   this lib and instead look for it at its absolute path.
+    --   When installing the .dylibs (see target.mk), we'll change that path to
+    --   point to the place they are installed. Therefore, we won't have to set
+    --	 up DYLD_LIBRARY_PATH specifically for ghc.
+    -----------------------------------------------------------------------------
+
+    let output_fn = case o_file of { Just s -> s; Nothing -> "a.out"; }
+
+    pwd <- getCurrentDirectory
+    SysTools.runLink dflags
+	 ([ SysTools.Option verb
+	  , SysTools.Option "-dynamiclib"
+	  , SysTools.Option "-o"
+	  , SysTools.FileOption "" output_fn
+	  ]
+	 ++ map SysTools.Option (
+	    md_c_flags
+	 ++ o_files
+	 ++ [ "-undefined", "dynamic_lookup", "-single_module", "-Wl,-macosx_version_min","-Wl,10.3", "-install_name " ++ (pwd `joinFileName` output_fn) ]
+	 ++ extra_ld_inputs
+	 ++ lib_path_opts
+	 ++ extra_ld_opts
+	 ++ pkg_lib_path_opts
+	 ++ pkg_link_opts
+	))
+#else
+    -----------------------------------------------------------------------------
+    -- Making a DSO
+    -----------------------------------------------------------------------------
+
+    let output_fn = case o_file of { Just s -> s; Nothing -> "a.out"; }
+
+    SysTools.runLink dflags
 	 ([ SysTools.Option verb
 	  , SysTools.Option "-o"
 	  , SysTools.FileOption "" output_fn
@@ -1322,18 +1458,14 @@ doMkDLL dflags o_files dep_packages = do
 	 ++ map SysTools.Option (
 	    md_c_flags
 	 ++ o_files
-	 ++ extra_os
-	 ++ [ "--target=i386-mingw32" ]
+	 ++ [ "-shared", "-Wl,-Bsymbolic" ] -- we need symbolic linking to resolve non-PIC intra-package-relocations
 	 ++ extra_ld_inputs
 	 ++ lib_path_opts
 	 ++ extra_ld_opts
 	 ++ pkg_lib_path_opts
 	 ++ pkg_link_opts
-         ++ (if "--def" `elem` (concatMap words extra_ld_opts)
-	       then [ "" ]
-               else [ "--export-all" ])
 	))
-
+#endif
 -- -----------------------------------------------------------------------------
 -- Running CPP
 
@@ -1414,8 +1546,6 @@ hscNextPhase dflags other hsc_lang =
 
 
 hscMaybeAdjustTarget :: DynFlags -> Phase -> HscSource -> HscTarget -> HscTarget
-hscMaybeAdjustTarget dflags stop HsBootFile current_hsc_lang 
-  = HscNothing		-- No output (other than Foo.hi-boot) for hs-boot files
 hscMaybeAdjustTarget dflags stop other current_hsc_lang 
   = hsc_lang 
   where
