@@ -1,4 +1,11 @@
 {-# OPTIONS -fno-warn-incomplete-patterns -optc-DNON_POSIX_SOURCE #-}
+{-# OPTIONS -w #-}
+-- The above warning supression flag is a temporary kludge.
+-- While working on this module you are encouraged to remove it and fix
+-- any warnings in the module. See
+--     http://hackage.haskell.org/trac/ghc/wiki/Commentary/CodingStyle#Warnings
+-- for details
+
 -----------------------------------------------------------------------------
 --
 -- GHC Driver program
@@ -13,40 +20,43 @@ module Main (main) where
 
 -- The official GHC API
 import qualified GHC
-import GHC		( Session, DynFlags(..), GhcMode(..), HscTarget(..),
+import GHC		( Session, DynFlags(..), HscTarget(..), 
+                          GhcMode(..), GhcLink(..),
 			  LoadHowMuch(..), dopt, DynFlag(..) )
 import CmdLineParser
 
 -- Implementations of the various modes (--show-iface, mkdependHS. etc.)
-import MkIface		( showIface )
+import LoadIface	( showIface )
+import HscMain          ( newHscEnv )
 import DriverPipeline	( oneShot, compileFile )
 import DriverMkDepend	( doMkDependHS )
 #ifdef GHCI
-import InteractiveUI	( ghciWelcomeMsg, interactiveUI )
+import InteractiveUI	( interactiveUI, ghciWelcomeMsg )
 #endif
 
 -- Various other random stuff that we need
-import Config		( cProjectVersion, cBooterVersion, cProjectName )
-import Packages		( dumpPackages, initPackages )
+import Config
+import Packages		( dumpPackages )
 import DriverPhases	( Phase(..), isSourceFilename, anyHsc,
 			  startPhase, isHaskellSrcFilename )
 import StaticFlags
-import DynFlags         ( defaultDynFlags )
+import DynFlags
 import BasicTypes	( failed )
-import ErrUtils		( Message, debugTraceMsg, putMsg )
+import ErrUtils		( putMsg )
 import FastString	( getFastStringTable, isZEncoded, hasZEncoding )
 import Outputable
 import Util
 import Panic
 
 -- Standard Haskell libraries
-import EXCEPTION	( throwDyn )
-import IO
-import Directory	( doesDirectoryExist )
-import System		( getArgs, exitWith, ExitCode(..) )
-import Monad
-import List
-import Maybe
+import Control.Exception ( throwDyn )
+import System.IO
+import System.Directory	( doesDirectoryExist )
+import System.Environment
+import System.Exit
+import Control.Monad
+import Data.List
+import Data.Maybe
 
 -----------------------------------------------------------------------------
 -- ToDo:
@@ -67,7 +77,7 @@ main =
   argv0 <- getArgs
 
   let
-        (minusB_args, argv1) = partition (prefixMatch "-B") argv0
+        (minusB_args, argv1) = partition ("-B" `isPrefixOf`) argv0
         mbMinusB | null minusB_args = Nothing
                  | otherwise = Just (drop 2 (last minusB_args))
 
@@ -82,35 +92,37 @@ main =
   -- to find out what version of GHC it's using before package.conf
   -- exists, so starting the session fails.
   case cli_mode of
-    ShowVersion     -> do showVersion
-                          exitWith ExitSuccess
-    ShowNumVersion  -> do putStrLn cProjectVersion
-                          exitWith ExitSuccess
-    _               -> return ()
-
-  let mode = case cli_mode of
-		DoInteractive	-> Interactive
-		DoEval _	-> Interactive
-		DoMake		-> BatchCompile
-		DoMkDependHS	-> MkDepend
-		_		-> OneShot
+    ShowInfo                -> do showInfo
+                                  exitWith ExitSuccess
+    ShowSupportedLanguages  -> do showSupportedLanguages
+                                  exitWith ExitSuccess
+    ShowVersion             -> do showVersion
+                                  exitWith ExitSuccess
+    ShowNumVersion          -> do putStrLn cProjectVersion
+                                  exitWith ExitSuccess
+    _                       -> return ()
 
   -- start our GHC session
-  session <- GHC.newSession mode mbMinusB
+  session <- GHC.newSession mbMinusB
 
   dflags0 <- GHC.getSessionDynFlags session
 
-  -- set the default HscTarget.  The HscTarget can be further
-  -- adjusted on a module by module basis, using only the -fvia-C and
-  -- -fasm flags.  If the default HscTarget is not HscC or HscAsm,
-  -- -fvia-C and -fasm have no effect.
-  let lang = case cli_mode of 
-		 DoInteractive  -> HscInterpreted
-		 DoEval _	-> HscInterpreted
-		 _other		-> hscTarget dflags0
+  -- set the default GhcMode, HscTarget and GhcLink.  The HscTarget
+  -- can be further adjusted on a module by module basis, using only
+  -- the -fvia-C and -fasm flags.  If the default HscTarget is not
+  -- HscC or HscAsm, -fvia-C and -fasm have no effect.
+  let dflt_target = hscTarget dflags0
+      (mode, lang, link)
+         = case cli_mode of
+		DoInteractive	-> (CompManager, HscInterpreted, LinkInMemory)
+		DoEval _	-> (CompManager, HscInterpreted, LinkInMemory)
+		DoMake		-> (CompManager, dflt_target,    LinkBinary)
+		DoMkDependHS	-> (MkDepend,    dflt_target,    LinkBinary)
+		_		-> (OneShot,     dflt_target,    LinkBinary)
 
-  let dflags1 = dflags0{ ghcMode = mode,
-                  	 hscTarget  = lang,
+  let dflags1 = dflags0{ ghcMode   = mode,
+                  	 hscTarget = lang,
+                         ghcLink   = link,
 			 -- leave out hscOutName for now
 	                 hscOutName = panic "Main.main:hscOutName not set",
 		  	 verbosity = case cli_mode of
@@ -125,7 +137,6 @@ main =
 	-- make sure we clean up after ourselves
   GHC.defaultCleanupHandler dflags $ do
 
-	-- Display banner
   showBanner cli_mode dflags
 
   -- we've finished manipulating the DynFlags, update the session
@@ -153,18 +164,21 @@ main =
 	---------------- Final sanity checking -----------
   checkOptions cli_mode dflags srcs objs
 
-	---------------- Do the business -----------
+  ---------------- Do the business -----------
+  let alreadyHandled = panic (show cli_mode ++
+                              " should already have been handled")
   case cli_mode of
-    ShowUsage       -> showGhcUsage dflags cli_mode
-    PrintLibdir     -> putStrLn (topDir dflags)
-    ShowVersion     -> panic "ShowVersion should already have been handled"
-    ShowNumVersion  -> panic "ShowNumVersion should already have been handled"
-    ShowInterface f -> showIface f
-    DoMake          -> doMake session srcs
-    DoMkDependHS    -> doMkDependHS session (map fst srcs)
-    StopBefore p    -> oneShot dflags p srcs
-    DoInteractive   -> interactiveUI session srcs Nothing
-    DoEval expr     -> interactiveUI session srcs (Just expr)
+    ShowUsage              -> showGhcUsage dflags cli_mode
+    PrintLibdir            -> putStrLn (topDir dflags)
+    ShowSupportedLanguages -> alreadyHandled
+    ShowVersion            -> alreadyHandled
+    ShowNumVersion         -> alreadyHandled
+    ShowInterface f        -> doShowIface dflags f
+    DoMake                 -> doMake session srcs
+    DoMkDependHS           -> doMkDependHS session (map fst srcs)
+    StopBefore p           -> oneShot dflags p srcs
+    DoInteractive          -> interactiveUI session srcs Nothing
+    DoEval expr            -> interactiveUI session srcs (Just expr)
 
   dumpFinalStats dflags
   exitWith ExitSuccess
@@ -222,11 +236,15 @@ checkOptions cli_mode dflags srcs objs = do
    let unknown_opts = [ f | (f@('-':_), _) <- srcs ]
    when (notNull unknown_opts) (unknownFlagsErr unknown_opts)
 
+   when (notNull (filter isRTSWay (wayNames dflags))
+         && isInterpretiveMode cli_mode) $
+        putStrLn ("Warning: -debug, -threaded and -ticky are ignored by GHCi")
+
 	-- -prof and --interactive are not a good combination
-   when (notNull (filter (/= WayThreaded) (wayNames dflags))
+   when (notNull (filter (not . isRTSWay) (wayNames dflags))
          && isInterpretiveMode cli_mode) $
       do throwDyn (UsageError 
-                   "--interactive can't be used with -prof, -ticky, or -unreg.")
+                   "--interactive can't be used with -prof or -unreg.")
 	-- -ohi sanity check
    if (isJust (outputHi dflags) && 
       (isCompManagerMode cli_mode || srcs `lengthExceeds` 1))
@@ -285,17 +303,19 @@ verifyOutputFiles dflags = do
 -- GHC modes of operation
 
 data CmdLineMode
-  = ShowUsage			-- ghc -?
-  | PrintLibdir			-- ghc --print-libdir
-  | ShowVersion			-- ghc -V/--version
-  | ShowNumVersion		-- ghc --numeric-version
-  | ShowInterface String	-- ghc --show-iface
-  | DoMkDependHS		-- ghc -M
-  | StopBefore Phase		-- ghc -E | -C | -S
-				-- StopBefore StopLn is the default
-  | DoMake			-- ghc --make
-  | DoInteractive		-- ghc --interactive
-  | DoEval String		-- ghc -e
+  = ShowUsage               -- ghc -?
+  | PrintLibdir             -- ghc --print-libdir
+  | ShowInfo                -- ghc --info
+  | ShowSupportedLanguages  -- ghc --supported-languages
+  | ShowVersion             -- ghc -V/--version
+  | ShowNumVersion          -- ghc --numeric-version
+  | ShowInterface String    -- ghc --show-iface
+  | DoMkDependHS            -- ghc -M
+  | StopBefore Phase        -- ghc -E | -C | -S
+                            -- StopBefore StopLn is the default
+  | DoMake                  -- ghc --make
+  | DoInteractive           -- ghc --interactive
+  | DoEval String           -- ghc -e
   deriving (Show)
 
 isInteractiveMode, isInterpretiveMode     :: CmdLineMode -> Bool
@@ -344,12 +364,14 @@ type ModeM a = CmdLineP (CmdLineMode, String, [String]) a
 mode_flags :: [(String, OptKind (CmdLineP (CmdLineMode, String, [String])))]
 mode_flags =
   [  ------- help / version ----------------------------------------------
-     ( "?"    		 , PassFlag (setMode ShowUsage))
-  ,  ( "-help"	 	 , PassFlag (setMode ShowUsage))
-  ,  ( "-print-libdir"   , PassFlag (setMode PrintLibdir))
-  ,  ( "V"	 	 , PassFlag (setMode ShowVersion))
-  ,  ( "-version"	 , PassFlag (setMode ShowVersion))
-  ,  ( "-numeric-version", PassFlag (setMode ShowNumVersion))
+     ( "?"                   , PassFlag (setMode ShowUsage))
+  ,  ( "-help"               , PassFlag (setMode ShowUsage))
+  ,  ( "-print-libdir"       , PassFlag (setMode PrintLibdir))
+  ,  ( "V"                   , PassFlag (setMode ShowVersion))
+  ,  ( "-version"            , PassFlag (setMode ShowVersion))
+  ,  ( "-numeric-version"    , PassFlag (setMode ShowNumVersion))
+  ,  ( "-info"               , PassFlag (setMode ShowInfo))
+  ,  ( "-supported-languages", PassFlag (setMode ShowSupportedLanguages))
 
       ------- interfaces ----------------------------------------------------
   ,  ( "-show-iface"     , HasArg (\f -> setMode (ShowInterface f)
@@ -408,30 +430,48 @@ doMake sess srcs  = do
     when (failed ok_flag) (exitWith (ExitFailure 1))
     return ()
 
+
+-- ---------------------------------------------------------------------------
+-- --show-iface mode
+
+doShowIface :: DynFlags -> FilePath -> IO ()
+doShowIface dflags file = do
+  hsc_env <- newHscEnv dflags
+  showIface hsc_env file
+
 -- ---------------------------------------------------------------------------
 -- Various banners and verbosity output.
 
 showBanner :: CmdLineMode -> DynFlags -> IO ()
 showBanner cli_mode dflags = do
    let verb = verbosity dflags
-	-- Show the GHCi banner
-#  ifdef GHCI
-   when (isInteractiveMode cli_mode && verb >= 1) $
-      hPutStrLn stdout ghciWelcomeMsg
-#  endif
 
-	-- Display details of the configuration in verbose mode
-   when (not (isInteractiveMode cli_mode) && verb >= 2) $
-	do hPutStr stderr "Glasgow Haskell Compiler, Version "
- 	   hPutStr stderr cProjectVersion
-	   hPutStr stderr ", for Haskell 98, compiled by GHC version "
 #ifdef GHCI
-	   -- GHCI is only set when we are bootstrapping...
- 	   hPutStrLn stderr cProjectVersion
-#else
-	   hPutStrLn stderr cBooterVersion
+   -- Show the GHCi banner
+   when (isInteractiveMode cli_mode && verb >= 1) $ putStrLn ghciWelcomeMsg
 #endif
 
+   -- Display details of the configuration in verbose mode
+   when (verb >= 2) $
+    do hPutStr stderr "Glasgow Haskell Compiler, Version "
+       hPutStr stderr cProjectVersion
+       hPutStr stderr ", for Haskell 98, stage "
+       hPutStr stderr cStage
+       hPutStr stderr " booted by GHC version "
+       hPutStrLn stderr cBooterVersion
+
+-- We print out a Read-friendly string, but a prettier one than the
+-- Show instance gives us
+showInfo :: IO ()
+showInfo = do
+    let sq x = " [" ++ x ++ "\n ]"
+    putStrLn $ sq $ concat $ intersperse "\n ," $ map show compilerInfo
+    exitWith ExitSuccess
+
+showSupportedLanguages :: IO ()
+showSupportedLanguages = do mapM_ putStrLn supportedLanguages
+                            exitWith ExitSuccess
+
 showVersion :: IO ()
 showVersion = do
   putStrLn (cProjectName ++ ", version " ++ cProjectVersion)
@@ -439,8 +479,8 @@ showVersion = do
 
 showGhcUsage dflags cli_mode = do 
   let usage_path 
-	| DoInteractive <- cli_mode = ghcUsagePath dflags
-	| otherwise		    = ghciUsagePath dflags
+	| DoInteractive <- cli_mode = ghciUsagePath dflags
+	| otherwise		    = ghcUsagePath dflags
   usage <- readFile usage_path
   dump usage
   exitWith ExitSuccess
