@@ -1,4 +1,11 @@
 {-# OPTIONS -fglasgow-exts -cpp #-}
+{-# OPTIONS -w #-}
+-- The above warning supression flag is a temporary kludge.
+-- While working on this module you are encouraged to remove it and fix
+-- any warnings in the module. See
+--     http://hackage.haskell.org/trac/ghc/wiki/Commentary/CodingStyle#Warnings
+-- for details
+
 module ParsePkgConf( loadPackageConfig ) where
 
 #include "HsVersions.h"
@@ -12,7 +19,7 @@ import ErrUtils  ( mkLocMessage )
 import SrcLoc
 import Outputable
 import Panic     ( GhcException(..) )
-import EXCEPTION ( throwDyn )
+import Control.Exception ( throwDyn )
 #if __GLASGOW_HASKELL__ >= 503
 import Data.Array
 #else
@@ -26,7 +33,12 @@ import GlaExts
 
 -- parser produced by Happy Version 1.16
 
-newtype HappyAbsSyn  = HappyAbsSyn (() -> ())
+newtype HappyAbsSyn  = HappyAbsSyn HappyAny
+#if __GLASGOW_HASKELL__ >= 607
+type HappyAny = GHC.Exts.Any
+#else
+type HappyAny = forall a . a
+#endif
 happyIn4 :: ([ PackageConfig ]) -> (HappyAbsSyn )
 happyIn4 x = unsafeCoerce# x
 {-# INLINE happyIn4 #-}
@@ -112,6 +124,7 @@ happyOutTok :: (HappyAbsSyn ) -> Located Token
 happyOutTok x = unsafeCoerce# x
 {-# INLINE happyOutTok #-}
 
+
 happyActOffsets :: HappyAddr
 happyActOffsets = HappyA# "\x4c\x00\x48\x00\x4a\x00\x45\x00\x0a\x00\x1c\x00\x00\x00\x00\x00\x47\x00\x46\x00\x00\x00\x44\x00\x00\x00\x16\x00\x00\x00\x3f\x00\x03\x00\x00\x00\x43\x00\x00\x00\x00\x00\x00\x00\x00\x00\x0c\x00\x01\x00\x00\x00\x42\x00\x00\x00\x41\x00\x40\x00\x3e\x00\x00\x00\x36\x00\x3c\x00\x3a\x00\x00\x00\x00\x00\x38\x00\x39\x00\x35\x00\x00\x00\x00\x00\x37\x00\x34\x00\x33\x00\x32\x00\x31\x00\x30\x00\x2f\x00\x00\x00\x2e\x00\x2b\x00\x2d\x00\x09\x00\x2c\x00\x00\x00\x2a\x00\x26\x00\x25\x00\x20\x00\x00\x00\x00\x00\x29\x00\x27\x00\x0d\x00\x00\x00\x00\x00"#
 
@@ -487,6 +500,7 @@ loadPackageConfig conf_filename = do
 	POk _ pkg_details -> do
 	    return pkg_details
 {-# LINE 1 "GenericTemplate.hs" #-}
+{-# LINE 1 "GenericTemplate.hs" #-}
 {-# LINE 1 "<built-in>" #-}
 {-# LINE 1 "<command line>" #-}
 {-# LINE 1 "GenericTemplate.hs" #-}
