@@ -1,3 +1,10 @@
+{-# OPTIONS -w #-}
+-- The above warning supression flag is a temporary kludge.
+-- While working on this module you are encouraged to remove it and fix
+-- any warnings in the module. See
+--     http://hackage.haskell.org/trac/ghc/wiki/Commentary/CodingStyle#Warnings
+-- for details
+
 -----------------------------------------------------------------------------
 --
 -- Static flags
@@ -12,9 +19,10 @@
 module StaticFlags (
 	parseStaticFlags,
 	staticFlags,
+        initStaticOpts,
 
 	-- Ways
-	WayName(..), v_Ways, v_Build_tag, v_RTS_Build_tag,
+	WayName(..), v_Ways, v_Build_tag, v_RTS_Build_tag, isRTSWay,
 
 	-- Output style options
 	opt_PprUserLength,
@@ -27,6 +35,9 @@ module StaticFlags (
 	opt_SccProfilingOn,
 	opt_DoTickyProfiling,
 
+        -- Hpc opts
+	opt_Hpc,
+
 	-- language opts
 	opt_DictsStrict,
 	opt_IrrefutableTuples,
@@ -37,9 +48,7 @@ module StaticFlags (
 	-- optimisation opts
 	opt_NoMethodSharing, 
 	opt_NoStateHack,
-	opt_LiberateCaseThreshold,
 	opt_CprOff,
-	opt_RulesOff,
 	opt_SimplNoPreInlining,
 	opt_SimplExcessPrecision,
 	opt_MaxWorkerArgs,
@@ -52,6 +61,11 @@ module StaticFlags (
 	opt_UF_UpdateInPlace,
 	opt_UF_DearOp,
 
+	-- Related to linking
+	opt_PIC,
+	opt_Static,
+	opt_HardwireLibPaths,
+
 	-- misc opts
 	opt_IgnoreDotGhci,
 	opt_ErrorSpans,
@@ -59,41 +73,41 @@ module StaticFlags (
 	opt_HiVersion,
 	opt_HistorySize,
 	opt_OmitBlackHoling,
-	opt_Static,
 	opt_Unregisterised,
 	opt_EmitExternalCore,
-	opt_PIC,
 	v_Ld_inputs,
+	tablesNextToCode
   ) where
 
 #include "HsVersions.h"
 
-import Util		( consIORef )
 import CmdLineParser
-import Config		( cProjectVersionInt, cProjectPatchLevel,
-			  cGhcUnregisterised )
+import Config
 import FastString	( FastString, mkFastString )
 import Util
 import Maybes		( firstJust )
-import Panic		( GhcException(..), ghcError )
+import Panic
 
-import EXCEPTION	( throwDyn )
-import DATA_IOREF
-import UNSAFE_IO	( unsafePerformIO )
-import Monad		( when )
-import Char		( isDigit )
-import Data.List	( sort, intersperse, nub )
+import Control.Exception ( throwDyn )
+import Data.IORef
+import System.IO.Unsafe	( unsafePerformIO )
+import Control.Monad	( when )
+import Data.Char	( isDigit )
+import Data.List
 
 -----------------------------------------------------------------------------
 -- Static flags
 
 parseStaticFlags :: [String] -> IO [String]
 parseStaticFlags args = do
+  ready <- readIORef v_opt_C_ready
+  when ready $ throwDyn (ProgramError "Too late for parseStaticFlags: call it before newSession")
+
   (leftover, errs) <- processArgs static_flags args
   when (not (null errs)) $ throwDyn (UsageError (unlines errs))
 
     -- deal with the way flags: the way (eg. prof) gives rise to
-    -- futher flags, some of which might be static.
+    -- further flags, some of which might be static.
   way_flags <- findBuildTag
 
     -- if we're unregisterised, add some more flags
@@ -101,23 +115,43 @@ parseStaticFlags args = do
 		  | otherwise = []
 
   (more_leftover, errs) <- processArgs static_flags (unreg_flags ++ way_flags)
+
+    -- see sanity code in staticOpts
+  writeIORef v_opt_C_ready True
+
+    -- TABLES_NEXT_TO_CODE affects the info table layout.
+    -- Be careful to do this *after* all processArgs,
+    -- because evaluating tablesNextToCode involves looking at the global
+    -- static flags.  Those pesky global variables...
+  let cg_flags | tablesNextToCode = ["-optc-DTABLES_NEXT_TO_CODE"]
+	       | otherwise	  = []
+
+    -- HACK: -fexcess-precision is both a static and a dynamic flag.  If
+    -- the static flag parser has slurped it, we must return it as a 
+    -- leftover too.  ToDo: make -fexcess-precision dynamic only.
+  let excess_prec | opt_SimplExcessPrecision = ["-fexcess-precision"]
+                  | otherwise                = []
+
   when (not (null errs)) $ ghcError (UsageError (unlines errs))
+  return (excess_prec++cg_flags++more_leftover++leftover)
 
-     -- urgh.  -fexcess-precision is both a static flag and a dynamic
-     -- flag.  See opt_SimplExcessPrecision below.  if we let the static
-     -- flag parser grab it, then we'll never get the dynamic version,
-     -- so we have to duplicate it.
-  let even_more | "-fexcess-precision" `elem` args = ["-fexcess-precision"]
-		| otherwise                        = []
+initStaticOpts :: IO ()
+initStaticOpts = writeIORef v_opt_C_ready True
 
-  return (even_more++more_leftover++leftover)
+static_flags :: [(String, OptKind IO)]
+-- All the static flags should appear in this list.  It describes how each
+-- static flag should be processed.  Two main purposes:
+-- (a) if a command-line flag doesn't appear in the list, GHC can complain
+-- (b) a command-line flag may remove, or add, other flags; e.g. the "-fno-X" things
+--
+-- The common (PassFlag addOpt) action puts the static flag into the bunch of
+-- things that are searched up by the top-level definitions like
+--	opt_foo = lookUp FSLIT("-dfoo")
 
-
--- note that ordering is important in the following list: any flag which
+-- Note that ordering is important in the following list: any flag which
 -- is a prefix flag (i.e. HasArg, Prefix, OptPrefix, AnySuffix) will override
 -- flags further down the list with the same prefix.
 
-static_flags :: [(String, OptKind IO)]
 static_flags = [
 	------- GHCi -------------------------------------------------------
      ( "ignore-dot-ghci", PassFlag addOpt )
@@ -136,7 +170,6 @@ static_flags = [
  	-- ToDo: user ways
 
 	------ Debugging ----------------------------------------------------
-  ,  ( "dppr-noprags",     PassFlag addOpt )
   ,  ( "dppr-debug",       PassFlag addOpt )
   ,  ( "dppr-user-length", AnySuffix addOpt )
       -- rest of the debugging flags are dynamic
@@ -188,7 +221,12 @@ lookup_str       :: String -> Maybe String
 -- holds the static opts while they're being collected, before
 -- being unsafely read by unpacked_static_opts below.
 GLOBAL_VAR(v_opt_C, defaultStaticOpts, [String])
-staticFlags = unsafePerformIO (readIORef v_opt_C)
+GLOBAL_VAR(v_opt_C_ready, False, Bool)
+staticFlags = unsafePerformIO $ do
+  ready <- readIORef v_opt_C_ready
+  if (not ready)
+        then panic "Static flags have not been initialised!\n        Please call GHC.newSession or GHC.parseStaticFlags early enough."
+        else readIORef v_opt_C
 
 -- -static is the default
 defaultStaticOpts = ["-static"]
@@ -253,7 +291,10 @@ opt_AutoSccsOnAllToplevs	= lookUp  FSLIT("-fauto-sccs-on-all-toplevs")
 opt_AutoSccsOnExportedToplevs	= lookUp  FSLIT("-fauto-sccs-on-exported-toplevs")
 opt_AutoSccsOnIndividualCafs	= lookUp  FSLIT("-fauto-sccs-on-individual-cafs")
 opt_SccProfilingOn		= lookUp  FSLIT("-fscc-profiling")
-opt_DoTickyProfiling		= lookUp  FSLIT("-fticky-ticky")
+opt_DoTickyProfiling            = WayTicky `elem` (unsafePerformIO $ readIORef v_Ways)
+
+-- Hpc opts
+opt_Hpc				= lookUp FSLIT("-fhpc")  
 
 -- language opts
 opt_DictsStrict			= lookUp  FSLIT("-fdicts-strict")
@@ -265,9 +306,7 @@ opt_Flatten			= lookUp  FSLIT("-fflatten")
 opt_NoStateHack			= lookUp  FSLIT("-fno-state-hack")
 opt_NoMethodSharing		= lookUp  FSLIT("-fno-method-sharing")
 opt_CprOff			= lookUp  FSLIT("-fcpr-off")
-opt_RulesOff			= lookUp  FSLIT("-frules-off")
 	-- Switch off CPR analysis in the new demand analyser
-opt_LiberateCaseThreshold	= lookup_def_int "-fliberate-case-threshold" (10::Int)
 opt_MaxWorkerArgs		= lookup_def_int "-fmax-worker-args" (10::Int)
 
 opt_GranMacros			= lookUp  FSLIT("-fgransim")
@@ -291,14 +330,27 @@ opt_UF_UpdateInPlace		= lookUp  FSLIT("-funfolding-update-in-place")
 
 opt_UF_DearOp   = ( 4 :: Int)
 			
+#if darwin_TARGET_OS && x86_64_TARGET_ARCH
+opt_PIC                         = True
+#else
+opt_PIC                         = lookUp FSLIT("-fPIC")
+#endif
 opt_Static			= lookUp  FSLIT("-static")
+opt_HardwireLibPaths		= lookUp  FSLIT("-fhardwire-lib-paths")
 opt_Unregisterised		= lookUp  FSLIT("-funregisterised")
+
+-- Derived, not a real option.  Determines whether we will be compiling
+-- info tables that reside just before the entry code, or with an
+-- indirection to the entry code.  See TABLES_NEXT_TO_CODE in 
+-- includes/InfoTables.h.
+tablesNextToCode 		= not opt_Unregisterised
+		 		  && cGhcEnableTablesNextToCode == "YES"
+
 opt_EmitExternalCore		= lookUp  FSLIT("-fext-core")
 
 -- Include full span info in error messages, instead of just the start position.
 opt_ErrorSpans			= lookUp FSLIT("-ferror-spans")
 
-opt_PIC                         = lookUp FSLIT("-fPIC")
 
 -- object files and libraries to be linked in are collected here.
 -- ToDo: perhaps this could be done without a global, it wasn't obvious
@@ -311,14 +363,10 @@ isStaticFlag f =
 	"fauto-sccs-on-exported-toplevs",
 	"fauto-sccs-on-individual-cafs",
 	"fscc-profiling",
-	"fticky-ticky",
-	"fall-strict",
 	"fdicts-strict",
 	"firrefutable-tuples",
 	"fparallel",
 	"fflatten",
-	"fsemi-tagging",
-	"flet-no-escape",
 	"fgransim",
 	"fno-hi-version-check",
 	"dno-black-holing",
@@ -329,15 +377,15 @@ isStaticFlag f =
 	"fexcess-precision",
 	"funfolding-update-in-place",
 	"static",
+	"fhardwire-lib-paths",
 	"funregisterised",
 	"fext-core",
-	"frule-check",
-	"frules-off",
 	"fcpr-off",
 	"ferror-spans",
-	"fPIC"
+	"fPIC",
+	"fhpc"
 	]
-  || any (flip prefixMatch f) [
+  || any (`isPrefixOf` f) [
 	"fliberate-case-threshold",
 	"fmax-worker-args",
 	"fhistory-size",
@@ -371,20 +419,15 @@ decodeSize str
   | c == "G" || c == "g" = truncate (n * 1000 * 1000 * 1000)
   | otherwise            = throwDyn (CmdLineError ("can't decode size: " ++ str))
   where (m, c) = span pred str
-        n      = read m  :: Double
+        n      = readRational m
 	pred c = isDigit c || c == '.'
 
 
 -----------------------------------------------------------------------------
 -- RTS Hooks
 
-#if __GLASGOW_HASKELL__ >= 504
 foreign import ccall unsafe "setHeapSize"       setHeapSize       :: Int -> IO ()
 foreign import ccall unsafe "enableTimingStats" enableTimingStats :: IO ()
-#else
-foreign import "setHeapSize"       unsafe setHeapSize       :: Int -> IO ()
-foreign import "enableTimingStats" unsafe enableTimingStats :: IO ()
-#endif
 
 -----------------------------------------------------------------------------
 -- Ways
@@ -458,6 +501,7 @@ findBuildTag :: IO [String]  -- new options
 findBuildTag = do
   way_names <- readIORef v_Ways
   let ws = sort (nub way_names)
+
   if not (allowed_combination ws)
       then throwDyn (CmdLineError $
       		    "combination not supported: "  ++
@@ -472,6 +516,8 @@ findBuildTag = do
       	   writeIORef v_RTS_Build_tag rts_tag
       	   return (concat flags)
 
+
+
 mkBuildTag :: [Way] -> String
 mkBuildTag ways = concat (intersperse "_" (map wayTag ways))
 
@@ -480,6 +526,8 @@ lkupWay w =
 	Nothing -> error "findBuildTag"
 	Just details -> details
 
+isRTSWay = wayRTSOnly . lkupWay 
+
 data Way = Way {
   wayTag     :: String,
   wayRTSOnly :: Bool,
@@ -510,9 +558,8 @@ way_details =
 	, "-DPROFILING"
 	, "-optc-DPROFILING" ]),
 
-    (WayTicky, Way  "t" False "Ticky-ticky Profiling"  
-	[ "-fticky-ticky"
-	, "-DTICKY_TICKY"
+    (WayTicky, Way  "t" True "Ticky-ticky Profiling"  
+	[ "-DTICKY_TICKY"
 	, "-optc-DTICKY_TICKY" ]),
 
     (WayUnreg, Way  "u" False "Unregisterised" 
