@@ -5,7 +5,9 @@ module Distribution.Compat.Map (
    member, lookup, findWithDefault,
    empty,
    insert, insertWith,
+   update,
    union, unionWith, unions,
+   difference,
    elems, keys,
    fromList, fromListWith,
    toAscList
@@ -38,9 +40,22 @@ empty = emptyFM
 insert :: Ord k => k -> a -> Map k a -> Map k a
 insert k a m = addToFM m k a
 
+-- This might be able to use delFromFM, but I'm confused by the
+--     IF_NOT_GHC(delFromFM COMMA)
+-- in the Data.FiniteMap export list in ghc 6.2.
+delete :: Ord k => k -> Map k a -> Map k a
+delete k m = delListFromFM m [k]
+
 insertWith :: Ord k => (a -> a -> a) -> k -> a -> Map k a -> Map k a
 insertWith c k a m = addToFM_C (flip c) m k a
 
+update :: Ord k => (a -> Maybe a) -> k -> Map k a -> Map k a
+update f k m = case lookup k m of
+  Nothing -> m
+  Just a  -> case f a of
+    Nothing -> delete k m
+    Just a' -> insert k a' m
+
 union :: Ord k => Map k a -> Map k a -> Map k a
 union = flip plusFM
 
@@ -50,6 +65,10 @@ unionWith c l r = plusFM_C (flip c) r l
 unions :: Ord k => [Map k a] -> Map k a
 unions = foldl (flip plusFM) emptyFM
 
+difference :: Ord k => Map k a -> Map k b -> Map k a
+difference m1 m2 = delListFromFM m1 (keys m2)
+  -- minusFM wasn't polymorphic enough in GHC 6.2.x
+
 elems :: Map k a -> [a]
 elems = eltsFM
 
