@@ -8,6 +8,8 @@
 -- Stability   :  alpha
 -- Portability :  portable
 --
+-- Invokes the "Distribution.Compiler"s to build the library and
+-- executables in this package.
 
 {- Copyright (c) 2003-2005, Isaac Jones
 All rights reserved.
@@ -41,59 +43,89 @@ THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. -}
 
 module Distribution.Simple.Build (
-	build
+	build, makefile, initialBuildSteps
 #ifdef DEBUG        
         ,hunitTests
 #endif
   ) where
 
-import Distribution.Compiler	( Compiler(..), CompilerFlavor(..) )
+import Distribution.Simple.Compiler	( Compiler(..), CompilerFlavor(..) )
 import Distribution.PackageDescription 
 				( PackageDescription(..), BuildInfo(..),
 				  setupMessage, Executable(..), Library(..), 
                                   autogenModuleName )
 import Distribution.Package 	( PackageIdentifier(..), showPackageId )
-import Distribution.Setup	 (CopyDest(..), BuildFlags(..) )
-import Distribution.PreProcess  ( preprocessSources, PPSuffixHandler )
+import Distribution.Simple.Setup	( CopyDest(..), BuildFlags(..), 
+                                  MakefileFlags(..) )
+import Distribution.Simple.PreProcess  ( preprocessSources, PPSuffixHandler )
 import Distribution.Simple.LocalBuildInfo
-				( LocalBuildInfo(..), mkBinDir, mkBinDirRel,
-				  mkLibDir, mkLibDirRel, mkDataDir,mkDataDirRel,
-				  mkLibexecDir, mkLibexecDirRel )
+				( LocalBuildInfo(..),
+                                  InstallDirs(..), absoluteInstallDirs,
+                                  prefixRelativeInstallDirs )
 import Distribution.Simple.Configure
 				( localBuildInfoFile )
-import Distribution.Simple.Utils( die )
+import Distribution.Simple.Utils( createDirectoryIfMissingVerbose, die )
+import Distribution.System
 
-import Distribution.Compat.Directory
-				( createDirectoryIfMissing )
-import Distribution.Compat.FilePath
-				( joinFileName, pathSeparator )
+import System.FilePath          ( (</>), pathSeparator )
 
-import Data.Maybe		( maybeToList, fromJust )
-import Control.Monad 		( unless )
-import System.Directory		( getModificationTime, doesFileExist)
+import Data.Maybe		( maybeToList, fromJust, isNothing )
+import Control.Monad 		( unless, when )
+import System.Directory		( getModificationTime, doesFileExist )
 
 import qualified Distribution.Simple.GHC  as GHC
 import qualified Distribution.Simple.JHC  as JHC
--- import qualified Distribution.Simple.NHC  as NHC
+import qualified Distribution.Simple.NHC  as NHC
 import qualified Distribution.Simple.Hugs as Hugs
 
-#ifdef mingw32_HOST_OS
 import Distribution.PackageDescription (hasLibs)
-#endif
+import Distribution.Verbosity
 
 #ifdef DEBUG
-import HUnit (Test)
+import Test.HUnit (Test)
 #endif
 
 -- -----------------------------------------------------------------------------
--- Build the library
+-- |Build the libraries and executables in this package.
 
-build :: PackageDescription
-         -> LocalBuildInfo
-         -> BuildFlags
-         -> [ PPSuffixHandler ]
+build    :: PackageDescription  -- ^mostly information from the .cabal file
+         -> LocalBuildInfo -- ^Configuration information
+         -> BuildFlags -- ^Flags that the user passed to build
+         -> [ PPSuffixHandler ] -- ^preprocessors to run before compiling
          -> IO ()
-build pkg_descr lbi (BuildFlags verbose) suffixes = do
+build pkg_descr lbi flags suffixes = do
+  let verbosity = buildVerbose flags
+  initialBuildSteps pkg_descr lbi verbosity suffixes
+  setupMessage verbosity "Building" pkg_descr
+  case compilerFlavor (compiler lbi) of
+    GHC  -> GHC.build  pkg_descr lbi verbosity
+    JHC  -> JHC.build  pkg_descr lbi verbosity
+    Hugs -> Hugs.build pkg_descr lbi verbosity
+    NHC  -> NHC.build  pkg_descr lbi verbosity
+    _    -> die ("Building is not supported with this compiler.")
+
+makefile :: PackageDescription  -- ^mostly information from the .cabal file
+         -> LocalBuildInfo -- ^Configuration information
+         -> MakefileFlags -- ^Flags that the user passed to makefile
+         -> [ PPSuffixHandler ] -- ^preprocessors to run before compiling
+         -> IO ()
+makefile pkg_descr lbi flags suffixes = do
+  let verb = makefileVerbose flags
+  initialBuildSteps pkg_descr lbi verb suffixes
+  when (not (hasLibs pkg_descr)) $
+      die ("Makefile is only supported for libraries, currently.")
+  setupMessage verb "Generating Makefile" pkg_descr
+  case compilerFlavor (compiler lbi) of
+    GHC  -> GHC.makefile  pkg_descr lbi flags
+    _    -> die ("Generating a Makefile is not supported for this compiler.")
+
+
+initialBuildSteps :: PackageDescription  -- ^mostly information from the .cabal file
+                  -> LocalBuildInfo -- ^Configuration information
+                  -> Verbosity -- ^The verbosity to use
+                  -> [ PPSuffixHandler ] -- ^preprocessors to run before compiling
+                  -> IO ()
+initialBuildSteps pkg_descr lbi verbosity suffixes = do
   -- check that there's something to build
   let buildInfos =
           map libBuildInfo (maybeToList (library pkg_descr)) ++
@@ -102,19 +134,13 @@ build pkg_descr lbi (BuildFlags verbose) suffixes = do
     let name = showPackageId (package pkg_descr)
     die ("Package " ++ name ++ " can't be built on this system.")
 
-  createDirectoryIfMissing True (buildDir lbi)
+  createDirectoryIfMissingVerbose verbosity True (buildDir lbi)
 
   -- construct and write the Paths_<pkg>.hs file
-  createDirectoryIfMissing True (autogenModulesDir lbi)
+  createDirectoryIfMissingVerbose verbosity True (autogenModulesDir lbi)
   buildPathsModule pkg_descr lbi
 
-  preprocessSources pkg_descr lbi verbose suffixes
-  setupMessage "Building" pkg_descr
-  case compilerFlavor (compiler lbi) of
-   GHC  -> GHC.build  pkg_descr lbi verbose
-   JHC  -> JHC.build  pkg_descr lbi verbose
-   Hugs -> Hugs.build pkg_descr lbi verbose
-   _    -> die ("Building is not supported with this compiler.")
+  preprocessSources pkg_descr lbi False verbosity suffixes
 
 -- ------------------------------------------------------------
 -- * Building Paths_<pkg>.hs
@@ -122,18 +148,19 @@ build pkg_descr lbi (BuildFlags verbose) suffixes = do
 
 -- The directory in which we put auto-generated modules
 autogenModulesDir :: LocalBuildInfo -> String
-autogenModulesDir lbi = buildDir lbi `joinFileName` "autogen"
+autogenModulesDir lbi = buildDir lbi </> "autogen"
 
 buildPathsModule :: PackageDescription -> LocalBuildInfo -> IO ()
 buildPathsModule pkg_descr lbi =
    let pragmas
-	| absolute = ""
+	| absolute || isHugs = ""
 	| otherwise =
 	  "{-# OPTIONS_GHC -fffi #-}\n"++
 	  "{-# LANGUAGE ForeignFunctionInterface #-}\n"
 
        foreign_imports
 	| absolute = ""
+	| isHugs = "import System.Environment\n"
 	| otherwise =
 	  "import Foreign\n"++
 	  "import Foreign.C\n"++
@@ -150,11 +177,13 @@ buildPathsModule pkg_descr lbi =
 	foreign_imports++
 	"import Data.Version"++
 	"\n"++
+	"\nversion :: Version"++
 	"\nversion = " ++ show (pkgVersion (package pkg_descr))++
 	"\n"
 
        body
 	| absolute =
+	  "\nbindir, libdir, datadir, libexecdir :: FilePath\n"++
 	  "\nbindir     = " ++ show flat_bindir ++
 	  "\nlibdir     = " ++ show flat_libdir ++
 	  "\ndatadir    = " ++ show flat_datadir ++
@@ -169,10 +198,10 @@ buildPathsModule pkg_descr lbi =
 	  "getDataFileName :: FilePath -> IO FilePath\n"++
 	  "getDataFileName name = return (datadir ++ "++path_sep++" ++ name)\n"
 	| otherwise =
-	  "\nprefix        = " ++ show (prefix lbi) ++
+	  "\nprefix        = " ++ show flat_prefix ++
 	  "\nbindirrel     = " ++ show (fromJust flat_bindirrel) ++
-	  "\n"++
-	  "\ngetBinDir :: IO FilePath\n"++
+	  "\n\n"++
+	  "getBinDir :: IO FilePath\n"++
 	  "getBinDir = getPrefixDirRel bindirrel\n\n"++
 	  "getLibDir :: IO FilePath\n"++
 	  "getLibDir = "++mkGetDir flat_libdir flat_libdirrel++"\n\n"++
@@ -185,7 +214,9 @@ buildPathsModule pkg_descr lbi =
 	  "  dir <- getDataDir\n"++
 	  "  return (dir `joinFileName` name)\n"++
 	  "\n"++
-	  get_prefix_stuff
+	  get_prefix_stuff++
+	  "\n"++
+	  filename_stuff
    in do btime <- getModificationTime localBuildInfoFile
    	 exists <- doesFileExist paths_filepath
    	 ptime <- if exists
@@ -195,32 +226,51 @@ buildPathsModule pkg_descr lbi =
 	   then writeFile paths_filepath (header++body)
 	   else return ()
  where
-	flat_bindir        = mkBinDir pkg_descr lbi NoCopyDest
-	flat_bindirrel     = mkBinDirRel pkg_descr lbi NoCopyDest
-	flat_libdir        = mkLibDir pkg_descr lbi NoCopyDest
-	flat_libdirrel     = mkLibDirRel pkg_descr lbi NoCopyDest
-	flat_datadir       = mkDataDir pkg_descr lbi NoCopyDest
-	flat_datadirrel    = mkDataDirRel pkg_descr lbi NoCopyDest
-	flat_libexecdir    = mkLibexecDir pkg_descr lbi NoCopyDest
-	flat_libexecdirrel = mkLibexecDirRel pkg_descr lbi NoCopyDest
+	InstallDirs {
+          prefix     = flat_prefix,
+          bindir     = flat_bindir,
+          libdir     = flat_libdir,
+          datadir    = flat_datadir,
+          libexecdir = flat_libexecdir
+        } = absoluteInstallDirs pkg_descr lbi NoCopyDest
+        InstallDirs {
+          bindir     = flat_bindirrel,
+          libdir     = flat_libdirrel,
+          datadir    = flat_datadirrel,
+          libexecdir = flat_libexecdirrel,
+          progdir    = flat_progdirrel
+        } = prefixRelativeInstallDirs pkg_descr lbi
 	
-	mkGetDir dir (Just dirrel) = "getPrefixDirRel " ++ show dirrel
+	mkGetDir _   (Just dirrel) = "getPrefixDirRel " ++ show dirrel
 	mkGetDir dir Nothing       = "return " ++ show dir
 
-#if mingw32_HOST_OS
-	absolute = hasLibs pkg_descr || flat_bindirrel == Nothing
-#else
-	absolute = True
-#endif
+        -- In several cases we cannot make relocatable installations
+        absolute =
+             hasLibs pkg_descr        -- we can only make progs relocatable
+          || isNothing flat_bindirrel -- if the bin dir is an absolute path
+          || not (supportsRelocatableProgs (compilerFlavor (compiler lbi)))
+
+        supportsRelocatableProgs Hugs = True
+        supportsRelocatableProgs GHC  = case os of
+                           Windows _ -> True
+                           _         -> False
+        supportsRelocatableProgs _    = False
 
   	paths_modulename = autogenModuleName pkg_descr
 	paths_filename = paths_modulename ++ ".hs"
-	paths_filepath = autogenModulesDir lbi `joinFileName` paths_filename
+	paths_filepath = autogenModulesDir lbi </> paths_filename
+
+	isHugs = compilerFlavor (compiler lbi) == Hugs
+        get_prefix_stuff
+          | isHugs    = "progdirrel :: String\n"++
+                        "progdirrel = "++show (fromJust flat_progdirrel)++"\n\n"++
+                        get_prefix_hugs
+          | otherwise = get_prefix_win32
 
 	path_sep = show [pathSeparator]
 
-get_prefix_stuff :: String
-get_prefix_stuff =
+get_prefix_win32 :: String
+get_prefix_win32 =
   "getPrefixDirRel :: FilePath -> IO FilePath\n"++
   "getPrefixDirRel dirRel = do \n"++
   "  let len = (2048::Int) -- plenty, PATH_MAX is 512 under Win32.\n"++
@@ -232,17 +282,27 @@ get_prefix_stuff =
   "     else do exePath <- peekCString buf\n"++
   "             free buf\n"++
   "             let (bindir,_) = splitFileName exePath\n"++
-  "             return (prefixFromBinDir bindir bindirrel `joinFileName` dirRel)\n"++
-  "  where\n"++
-  "    prefixFromBinDir bindir path\n"++
-  "      | path == \".\" || path == \"\" = bindir\n"++
-  "      | otherwise                 = prefixFromBinDir bindir' path'\n"++
-  "      where\n"++
-  "        (bindir',_) = splitFileName bindir\n"++
-  "        (path',  _) = splitFileName path\n"++
+  "             return ((bindir `minusFileName` bindirrel) `joinFileName` dirRel)\n"++
   "\n"++
   "foreign import stdcall unsafe \"windows.h GetModuleFileNameA\"\n"++
-  "  getModuleFileName :: Ptr () -> CString -> Int -> IO Int32\n"++
+  "  getModuleFileName :: Ptr () -> CString -> Int -> IO Int32\n"
+
+get_prefix_hugs :: String
+get_prefix_hugs =
+  "getPrefixDirRel :: FilePath -> IO FilePath\n"++
+  "getPrefixDirRel dirRel = do\n"++
+  "  mainPath <- getProgName\n"++
+  "  let (progPath,_) = splitFileName mainPath\n"++
+  "  let (progdir,_) = splitFileName progPath\n"++
+  "  return ((progdir `minusFileName` progdirrel) `joinFileName` dirRel)\n"
+
+filename_stuff :: String
+filename_stuff =
+  "minusFileName :: FilePath -> String -> FilePath\n"++
+  "minusFileName dir \"\"     = dir\n"++
+  "minusFileName dir \".\"    = dir\n"++
+  "minusFileName dir suffix =\n"++
+  "  minusFileName (fst (splitFileName dir)) (fst (splitFileName suffix))\n"++
   "\n"++
   "joinFileName :: String -> String -> FilePath\n"++
   "joinFileName \"\"  fname = fname\n"++
@@ -252,6 +312,7 @@ get_prefix_stuff =
   "  | isPathSeparator (last dir) = dir++fname\n"++
   "  | otherwise                  = dir++pathSeparator:fname\n"++
   "\n"++
+  "splitFileName :: FilePath -> (String, String)\n"++
   "splitFileName p = (reverse (path2++drive), reverse fname)\n"++
   "  where\n"++
   "    (path,drive) = case p of\n"++
@@ -266,11 +327,14 @@ get_prefix_stuff =
   "      _                            -> path1\n"++
   "\n"++
   "pathSeparator :: Char\n"++
-  "pathSeparator = '\\\\'\n"++
+  (case os of
+       Windows _ -> "pathSeparator = '\\\\'\n"
+       _         -> "pathSeparator = '/'\n") ++
   "\n"++
   "isPathSeparator :: Char -> Bool\n"++
-  "isPathSeparator ch =\n"++
-  "  ch == '/' || ch == '\\\\'\n"
+  (case os of
+       Windows _ -> "isPathSeparator c = c == '/' || c == '\\\\'\n"
+       _         -> "isPathSeparator c = c == '/'\n")
 
 -- ------------------------------------------------------------
 -- * Testing
