@@ -42,34 +42,37 @@ THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. -}
 
 module Distribution.Simple.Utils (
-	die,
-	dieWithLocation,
-	warn,
-	rawSystemPath,
-        rawSystemVerbose,
-	rawSystemExit,
+        die,
+        dieWithLocation,
+        warn, notice, info, debug,
+        breaks,
+	wrapText,
+        rawSystemExit,
+        rawSystemStdout,
+	rawSystemStdout',
         maybeExit,
         xargs,
         matchesDescFile,
 	rawSystemPathExit,
         smartCopySources,
+        createDirectoryIfMissingVerbose,
         copyFileVerbose,
         copyDirectoryRecursiveVerbose,
         moduleToFilePath,
+        moduleToFilePath2,
         mkLibName,
         mkProfLibName,
+        mkSharedLibName,
         currentDir,
-	dirOf,
         dotToSep,
-	withTempFile,
 	findFile,
         defaultPackageDesc,
         findPackageDesc,
 	defaultHookedPackageDesc,
 	findHookedPackageDesc,
-        distPref,
-        haddockPref,
-        srcPref,
+        exeExtension,
+        objExtension,
+        dllExtension,
 #ifdef DEBUG
         hunitTests
 #endif
@@ -83,32 +86,56 @@ module Distribution.Simple.Utils (
 #endif
 #endif
 
-import Distribution.Compat.RawSystem (rawSystem)
-import Distribution.Compat.Exception (finally)
+import Control.Monad
+    ( when, filterM, unless, liftM2 )
+import Data.List
+    ( nub, unfoldr )
 
-import Control.Monad(when, filterM, unless)
-import Data.List (nub, unfoldr)
-import System.Environment (getProgName)
-import System.IO (hPutStrLn, stderr, hFlush, stdout)
-import System.IO.Error
+import System.Directory
+    ( getDirectoryContents, getCurrentDirectory, doesDirectoryExist
+    , doesFileExist, removeFile )
+import System.Environment
+    ( getProgName )
 import System.Exit
-#if (__GLASGOW_HASKELL__ || __HUGS__) && !(mingw32_HOST_OS || mingw32_TARGET_OS)
-import System.Posix.Internals (c_getpid)
-#endif
-
-import Distribution.Compat.FilePath
-	(splitFileName, splitFileExt, joinFileName, joinFileExt, joinPaths,
-	pathSeparator,splitFilePath)
-import System.Directory (getDirectoryContents, getCurrentDirectory
-			, doesDirectoryExist, doesFileExist, removeFile, getPermissions
-			, Permissions(executable))
+    ( exitWith, ExitCode(..) )
+import System.FilePath
+    ( takeDirectory, takeExtension, (</>), (<.>), pathSeparator )
+import System.IO
+    ( hPutStrLn, stderr, hFlush, stdout, openFile, IOMode(WriteMode) )
+import System.IO.Error
+    ( try )
 
 import Distribution.Compat.Directory
-           (copyFile, findExecutable, createDirectoryIfMissing,
-            getDirectoryContentsWithoutSpecial)
+    ( copyFile, findExecutable, createDirectoryIfMissing
+    , getDirectoryContentsWithoutSpecial)
+import Distribution.Compat.RawSystem
+    ( rawSystem )
+import Distribution.Compat.Exception
+    ( bracket )
+import Distribution.System
+    ( OS(..), os )
+import Distribution.Version
+    (showVersion)
+import Distribution.Package
+    (PackageIdentifier(..))
+
+#if __GLASGOW_HASKELL__ >= 604
+import Control.Exception (evaluate)
+import System.Process (runProcess, waitForProcess)
+#else
+import System.Cmd (system)
+#endif
+import System.IO (hClose)
+
+#if __GLASGOW_HASKELL__ >= 604
+import Distribution.Compat.TempFile (openTempFile)
+#else
+import Distribution.Compat.TempFile (withTempFile)
+#endif
+import Distribution.Verbosity
 
 #ifdef DEBUG
-import HUnit ((~:), (~=?), Test(..), assertEqual)
+import Test.HUnit ((~:), (~=?), Test(..), assertEqual)
 #endif
 
 -- ------------------------------------------------------------------------------- Utils for setup
@@ -124,51 +151,138 @@ die msg = do
   hPutStrLn stderr (pname ++ ": " ++ msg)
   exitWith (ExitFailure 1)
 
-warn :: String -> IO ()
-warn msg = do
-  hFlush stdout
-  pname <- getProgName
-  hPutStrLn stderr (pname ++ ": Warning: " ++ msg)
+-- | Non fatal conditions that may be indicative of an error or problem.
+--
+-- We display these at the 'normal' verbosity level.
+--
+warn :: Verbosity -> String -> IO ()
+warn verbosity msg = 
+  when (verbosity >= normal) $ do
+    hFlush stdout
+    hPutStrLn stderr ("Warning: " ++ msg)
+
+-- | Useful status messages.
+--
+-- We display these at the 'normal' verbosity level.
+--
+-- This is for the ordinary helpful status messages that users see. Just
+-- enough information to know that things are working but not floods of detail.
+--
+notice :: Verbosity -> String -> IO ()
+notice verbosity msg =
+  when (verbosity >= normal) $
+    putStrLn msg
+
+-- | More detail on the operation of some action.
+-- 
+-- We display these messages when the verbosity level is 'verbose'
+--
+info :: Verbosity -> String -> IO ()
+info verbosity msg =
+  when (verbosity >= verbose) $
+    putStrLn msg
+
+-- | Detailed internal debugging information
+--
+-- We display these messages when the verbosity level is 'deafening'
+--
+debug :: Verbosity -> String -> IO ()
+debug verbosity msg =
+  when (verbosity >= deafening) $
+    putStrLn msg
+
+-- -----------------------------------------------------------------------------
+-- Helper functions
+
+breaks :: (a -> Bool) -> [a] -> [[a]]
+breaks _ [] = []
+breaks f xs = case span f xs of
+                  (_, xs') ->
+                      case break f xs' of
+                          (v, xs'') ->
+                              v : breaks f xs''
+
+-- Wraps a list of words text to a list of lines of a particular width.
+wrapText :: Int -> [String] -> [String]
+wrapText width = map unwords . wrap 0 []
+  where wrap :: Int -> [String] -> [String] -> [[String]]
+        wrap 0   []   (w:ws)
+          | length w + 1 > width
+          = wrap (length w) [w] ws
+        wrap col line (w:ws)
+          | col + length w + 1 > width
+          = reverse line : wrap 0 [] (w:ws)
+        wrap col line (w:ws)
+          = let col' = col + length w + 1
+             in wrap col' (w:line) ws
+        wrap _ []   [] = []
+        wrap _ line [] = [reverse line]
 
 -- -----------------------------------------------------------------------------
 -- rawSystem variants
-rawSystemPath :: Int -> String -> [String] -> IO ExitCode
-rawSystemPath verbose prog args = do
-  r <- findExecutable prog
-  case r of
-    Nothing   -> die ("Cannot find: " ++ prog)
-    Just path -> rawSystemVerbose verbose path args
-
-rawSystemVerbose :: Int -> FilePath -> [String] -> IO ExitCode
-rawSystemVerbose verbose prog args = do
-      when (verbose > 0) $
-        putStrLn (prog ++ concatMap (' ':) args)
-      e <- doesFileExist prog
-      if e
-         then do perms <- getPermissions prog
-                 if (executable perms)
-                    then rawSystem prog args
-                    else die ("Error: file is not executable: " ++ show prog)
-         else die ("Error: file does not exist: " ++ show prog)
-
 maybeExit :: IO ExitCode -> IO ()
 maybeExit cmd = do
   res <- cmd
-  if res /= ExitSuccess
-	then exitWith res  
-	else return ()
+  unless (res == ExitSuccess) $ exitWith res
+
+printRawCommandAndArgs :: Verbosity -> FilePath -> [String] -> IO ()
+printRawCommandAndArgs verbosity path args
+ | verbosity >= deafening = print (path, args)
+ | verbosity >= verbose   = putStrLn $ unwords (path : args)
+ | otherwise              = return ()
 
 -- Exit with the same exitcode if the subcommand fails
-rawSystemExit :: Int -> FilePath -> [String] -> IO ()
-rawSystemExit verbose path args = do
-  when (verbose > 0) $
-    putStrLn (path ++ concatMap (' ':) args)
+rawSystemExit :: Verbosity -> FilePath -> [String] -> IO ()
+rawSystemExit verbosity path args = do
+  printRawCommandAndArgs verbosity path args
   maybeExit $ rawSystem path args
 
 -- Exit with the same exitcode if the subcommand fails
-rawSystemPathExit :: Int -> String -> [String] -> IO ()
-rawSystemPathExit verbose prog args = do
-  maybeExit $ rawSystemPath verbose prog args
+rawSystemPathExit :: Verbosity -> String -> [String] -> IO ()
+rawSystemPathExit verbosity prog args = do
+  r <- findExecutable prog
+  case r of
+    Nothing   -> die ("Cannot find: " ++ prog)
+    Just path -> rawSystemExit verbosity path args
+
+-- Run a command and return its output
+rawSystemStdout :: Verbosity -> FilePath -> [String] -> IO String
+rawSystemStdout verbosity path args = do
+  (output, exitCode) <- rawSystemStdout' verbosity path args
+  unless (exitCode == ExitSuccess) $ exitWith exitCode
+  return output
+
+rawSystemStdout' :: Verbosity -> FilePath -> [String] -> IO (String, ExitCode)
+rawSystemStdout' verbosity path args = do
+  printRawCommandAndArgs verbosity path args
+
+#if __GLASGOW_HASKELL__ >= 604
+  -- TODO Ideally we'd use runInteractiveProcess and not have to make any
+  --      silly temp files, however it is not possible to only connect pipes
+  --      to a subset of the process's stdin/out/err. We really cannot
+  --      connect to all three since then we'd need threads to pull on stdout
+  --      and stderr simultaniously to avoid deadlock, and using threads like
+  --      that would not be portable to Hugs for example.
+  bracket (liftM2 (,) (openTempFile "." "tmp") (openFile devNull WriteMode))
+          -- We need to close tmpHandle or the file removal fails on Windows
+          (\((tmpName, tmpHandle), nullHandle) -> do
+             hClose tmpHandle
+             removeFile tmpName
+             hClose nullHandle)
+         $ \((tmpName, tmpHandle), nullHandle) -> do
+    cmdHandle <- runProcess path args Nothing Nothing
+                   Nothing (Just tmpHandle) (Just nullHandle)
+    exitCode <- waitForProcess cmdHandle
+    output <- readFile tmpName
+    evaluate (length output)
+    return (output, exitCode)
+#else
+  withTempFile "." "" $ \tmpName -> do
+    let quote name = "'" ++ name ++ "'"
+    exitCode <- system $ unwords (map quote (path:args)) ++ " >" ++ quote tmpName
+    output <- readFile tmpName
+    length output `seq` return (output, exitCode)
+#endif
 
 -- | Like the unix xargs program. Useful for when we've got very long command
 -- lines that might overflow an OS limit on command line length and so you
@@ -176,26 +290,20 @@ rawSystemPathExit verbose prog args = do
 --
 -- Use it with either of the rawSystem variants above. For example:
 -- 
--- > xargs (32*1024) (rawSystemPath verbose) prog fixedArgs bigArgs
+-- > xargs (32*1024) (rawSystemPathExit verbosity) prog fixedArgs bigArgs
 --
-xargs :: Int -> (FilePath -> [String] -> IO ExitCode)
-      -> FilePath -> [String] -> [String] -> IO ExitCode
-xargs maxSize rawSystem prog fixedArgs bigArgs =
+xargs :: Int -> ([String] -> IO ())
+      -> [String] -> [String] -> IO ()
+xargs maxSize rawSystemFun fixedArgs bigArgs =
   let fixedArgSize = sum (map length fixedArgs) + length fixedArgs
       chunkSize = maxSize - fixedArgSize
-      loop [] = return ExitSuccess
-      loop (args:remainingArgs) = do
-        status <- rawSystem prog (fixedArgs ++ args)
-        case status of
-          ExitSuccess -> loop remainingArgs
-          _           -> return status
-   in loop (chunks chunkSize bigArgs)
+   in mapM_ (rawSystemFun . (fixedArgs ++)) (chunks chunkSize bigArgs)
 
   where chunks len = unfoldr $ \s ->
           if null s then Nothing
                     else Just (chunk [] len s)
 
-        chunk acc len []     = (reverse acc,[])
+        chunk acc _   []     = (reverse acc,[])
         chunk acc len (s:ss)
           | len' < len = chunk (s:acc) (len-len'-1) ss
           | otherwise  = (reverse acc, s:ss)
@@ -232,10 +340,10 @@ moduleToFilePath2
     -> IO [(FilePath, FilePath)] -- ^locations and relative names
 moduleToFilePath2 locs mname possibleSuffixes
     = filterM exists $
-        [(loc, fname `joinFileExt` ext) | loc <- locs, ext <- possibleSuffixes]
+        [(loc, fname <.> ext) | loc <- locs, ext <- possibleSuffixes]
   where
     fname = dotToSep mname
-    exists (loc, relname) = doesFileExist (loc `joinFileName` relname)
+    exists (loc, relname) = doesFileExist (loc </> relname)
 
 -- |Get the possible file paths based on this module name.
 moduleToPossiblePaths :: FilePath -- ^search prefix
@@ -243,19 +351,19 @@ moduleToPossiblePaths :: FilePath -- ^search prefix
                       -> [String] -- ^possible suffixes
                       -> [FilePath]
 moduleToPossiblePaths searchPref s possibleSuffixes =
-  let fname = searchPref `joinFileName` (dotToSep s)
-  in [fname `joinFileExt` ext | ext <- possibleSuffixes]
+  let fname = searchPref </> (dotToSep s)
+  in [fname <.> ext | ext <- possibleSuffixes]
 
 findFile :: [FilePath]    -- ^search locations
          -> FilePath      -- ^File Name
          -> IO FilePath
 findFile prefPathsIn locPath = do
   let prefPaths = nub prefPathsIn -- ignore dups
-  paths <- filterM doesFileExist [prefPath `joinFileName` locPath | prefPath <- prefPaths]
+  paths <- filterM doesFileExist [prefPath </> locPath | prefPath <- prefPaths]
   case nub paths of -- also ignore dups, though above nub should fix this.
     [path] -> return path
     []     -> die (locPath ++ " doesn't exist")
-    paths  -> die (locPath ++ " is found in multiple places:" ++ unlines (map ((++) "    ") paths))
+    paths' -> die (locPath ++ " is found in multiple places:" ++ unlines (map ((++) "    ") paths'))
 
 dotToSep :: String -> String
 dotToSep = map dts
@@ -268,7 +376,7 @@ dotToSep = map dts
 -- the input search suffixes.  It copies the files into the target
 -- directory.
 
-smartCopySources :: Int      -- ^verbose
+smartCopySources :: Verbosity -- ^verbosity
             -> [FilePath] -- ^build prefix (location of objects)
             -> FilePath -- ^Target directory
             -> [String] -- ^Modules
@@ -276,19 +384,19 @@ smartCopySources :: Int      -- ^verbose
             -> Bool     -- ^Exit if no such modules
             -> Bool     -- ^Preserve directory structure
             -> IO ()
-smartCopySources verbose srcDirs targetDir sources searchSuffixes exitIfNone preserveDirs
-    = do createDirectoryIfMissing True targetDir
+smartCopySources verbosity srcDirs targetDir sources searchSuffixes exitIfNone preserveDirs
+    = do createDirectoryIfMissingVerbose verbosity True targetDir
          allLocations <- mapM moduleToFPErr sources
-         let copies = [(srcDir `joinFileName` name,
+         let copies = [(srcDir </> name,
                         if preserveDirs 
-                          then targetDir `joinFileName` srcDir `joinFileName` name
-                          else targetDir `joinFileName` name) |
+                          then targetDir </> srcDir </> name
+                          else targetDir </> name) |
                        (srcDir, name) <- concat allLocations]
 	 -- Create parent directories for everything:
-	 mapM_ (createDirectoryIfMissing True) $ nub $
-             [fst (splitFileName targetFile) | (_, targetFile) <- copies]
+	 mapM_ (createDirectoryIfMissingVerbose verbosity True) $ nub $
+             [takeDirectory targetFile | (_, targetFile) <- copies]
 	 -- Put sources into place:
-	 sequence_ [copyFileVerbose verbose srcFile destFile |
+	 sequence_ [copyFileVerbose verbosity srcFile destFile |
                     (srcFile, destFile) <- copies]
     where moduleToFPErr m
               = do p <- moduleToFilePath2 srcDirs m searchSuffixes
@@ -297,29 +405,33 @@ smartCopySources verbose srcDirs targetDir sources searchSuffixes exitIfNone pre
                                        ++ " with any suffix: " ++ (show searchSuffixes)))
                    return p
 
-copyFileVerbose :: Int -> FilePath -> FilePath -> IO ()
-copyFileVerbose verbose src dest = do
-  when (verbose > 0) $
-    putStrLn ("copy " ++ src ++ " to " ++ dest)
+createDirectoryIfMissingVerbose :: Verbosity -> Bool -> FilePath -> IO ()
+createDirectoryIfMissingVerbose verbosity parentsToo dir = do
+  let msgParents = if parentsToo then " (and its parents)" else ""
+  info verbosity ("Creating " ++ dir ++ msgParents)
+  createDirectoryIfMissing parentsToo dir
+
+copyFileVerbose :: Verbosity -> FilePath -> FilePath -> IO ()
+copyFileVerbose verbosity src dest = do
+  info verbosity ("copy " ++ src ++ " to " ++ dest)
   copyFile src dest
 
 -- adaptation of removeDirectoryRecursive
-copyDirectoryRecursiveVerbose :: Int -> FilePath -> FilePath -> IO ()
-copyDirectoryRecursiveVerbose verbose srcDir destDir = do
-  when (verbose > 0) $
-    putStrLn ("copy directory '" ++ srcDir ++ "' to '" ++ destDir ++ "'.")
+copyDirectoryRecursiveVerbose :: Verbosity -> FilePath -> FilePath -> IO ()
+copyDirectoryRecursiveVerbose verbosity srcDir destDir = do
+  info verbosity ("copy directory '" ++ srcDir ++ "' to '" ++ destDir ++ "'.")
   let aux src dest =
          let cp :: FilePath -> IO ()
-             cp f = let srcFile  = joinPaths src  f
-                        destFile = joinPaths dest f
-                    in  do success <- try (copyFileVerbose verbose srcFile destFile)
+             cp f = let srcFile  = src  </> f
+                        destFile = dest </> f
+                    in  do success <- try (copyFileVerbose verbosity srcFile destFile)
                            case success of
                               Left e  -> do isDir <- doesDirectoryExist srcFile
                                             -- If f is not a directory, re-throw the error
                                             unless isDir $ ioError e
                                             aux srcFile destFile
                               Right _ -> return ()
-         in  do createDirectoryIfMissing False dest
+         in  do createDirectoryIfMissingVerbose verbosity False dest
                 getDirectoryContentsWithoutSpecial src >>= mapM_ cp
    in aux srcDir destDir
 
@@ -332,59 +444,26 @@ copyDirectoryRecursiveVerbose verbose srcDir destDir = do
 currentDir :: FilePath
 currentDir = "."
 
-dirOf :: FilePath -> FilePath
-dirOf f = (\ (x, _, _) -> x) $ (splitFilePath f)
-
 mkLibName :: FilePath -- ^file Prefix
           -> String   -- ^library name.
           -> String
-mkLibName pref lib = pref `joinFileName` ("libHS" ++ lib ++ ".a")
+mkLibName pref lib = pref </> ("libHS" ++ lib ++ ".a")
 
 mkProfLibName :: FilePath -- ^file Prefix
               -> String   -- ^library name.
               -> String
 mkProfLibName pref lib = mkLibName pref (lib++"_p")
 
--- ------------------------------------------------------------
--- * Some Paths
--- ------------------------------------------------------------
-distPref :: FilePath
-distPref = "dist"
-
-srcPref :: FilePath
-srcPref = distPref `joinFileName` "src"
-
-haddockPref :: FilePath
-haddockPref = foldl1 joinPaths [distPref, "doc", "html"]
-
--- ------------------------------------------------------------
--- * temporary file names
--- ------------------------------------------------------------
-
--- use a temporary filename that doesn't already exist.
--- NB. *not* secure (we don't atomically lock the tmp file we get)
-withTempFile :: FilePath -> String -> (FilePath -> IO a) -> IO a
-withTempFile tmp_dir extn action
-  = do x <- getProcessID
-       findTempName x
-  where 
-    findTempName x
-      = do let filename = ("tmp" ++ show x) `joinFileExt` extn
-	       path = tmp_dir `joinFileName` filename
-  	   b  <- doesFileExist path
-	   if b then findTempName (x+1)
-		else action path `finally` try (removeFile path)
-
-#if mingw32_HOST_OS || mingw32_TARGET_OS
-foreign import ccall unsafe "_getpid" getProcessID :: IO Int
-		 -- XXX relies on Int == Int32 on Windows
-#else
-#if !(__GLASGOW_HASKELL__ || __HUGS__)
-foreign import ccall unsafe "getpid" c_getpid :: IO CPid
-#endif
-getProcessID :: IO Int
-getProcessID = c_getpid >>= return . fromIntegral
-#endif
+-- Implement proper name mangling for dynamical shared objects
+-- libHS<packagename>-<compilerFlavour><compilerVersion>
+-- e.g. libHSbase-2.1-ghc6.6.1.so
+mkSharedLibName :: FilePath        -- ^file Prefix
+              -> String            -- ^library name.
+              -> PackageIdentifier -- ^package identifier of the compiler
+              -> String
+mkSharedLibName pref lib (PackageIdentifier compilerName compilerVersion)
+  = pref </> ("libHS" ++ lib ++ "-" ++ compiler) <.> dllExtension
+  where compiler = compilerName ++ showVersion compilerVersion
 
 -- ------------------------------------------------------------
 -- * Finding the description file
@@ -399,8 +478,10 @@ cabalExt = "cabal"
 buildInfoExt  :: String
 buildInfoExt = "buildinfo"
 
+-- > matchesDescFile "blah.Cabal"
+-- > not $ matchesDescFile "blag.bleg"
 matchesDescFile :: FilePath -> Bool
-matchesDescFile p = (snd $ splitFileExt p) == cabalExt
+matchesDescFile p = (takeExtension p) == '.':cabalExt
                     || p == oldDescFile
 
 noDesc :: IO a
@@ -411,38 +492,41 @@ multiDesc l = die $ "Multiple description files found.  Please use only one of :
                       ++ show (filter (/= oldDescFile) l)
 
 -- |A list of possibly correct description files.  Should be pre-filtered.
-descriptionCheck :: [FilePath] -> IO FilePath
-descriptionCheck [] = noDesc
-descriptionCheck [x]
+descriptionCheck :: Verbosity -> [FilePath] -> IO FilePath
+descriptionCheck _ [] = noDesc
+descriptionCheck verbosity [x]
     | x == oldDescFile
-        = do warn $ "The filename \"Setup.description\" is deprecated, please move to <pkgname>." ++ cabalExt
+        = do warn verbosity $ "The filename \"Setup.description\" is deprecated, please move to <pkgname>." ++ cabalExt
              return x
     | matchesDescFile x = return x
     | otherwise = noDesc
-descriptionCheck [x,y]
+descriptionCheck verbosity [x,y]
     | x == oldDescFile
-        = do warn $ "The filename \"Setup.description\" is deprecated.  Please move out of the way. Using \""
+        = do warn verbosity $ "The filename \"Setup.description\" is deprecated.  Please move out of the way. Using \""
                   ++ y ++ "\""
              return y
     | y == oldDescFile
-        = do warn $ "The filename \"Setup.description\" is deprecated.  Please move out of the way. Using \""
+        = do warn verbosity $ "The filename \"Setup.description\" is deprecated.  Please move out of the way. Using \""
                   ++ x ++ "\""
              return x
 
     | otherwise = multiDesc [x,y]
-descriptionCheck l = multiDesc l
+descriptionCheck _ l = multiDesc l
 
 -- |Package description file (/pkgname/@.cabal@)
-defaultPackageDesc :: IO FilePath
-defaultPackageDesc = getCurrentDirectory >>= findPackageDesc
+defaultPackageDesc :: Verbosity -> IO FilePath
+defaultPackageDesc verbosity
+    = getCurrentDirectory >>= findPackageDesc verbosity
 
 -- |Find a package description file in the given directory.  Looks for
 -- @.cabal@ files.
-findPackageDesc :: FilePath    -- ^Where to look
+findPackageDesc :: Verbosity   -- ^Verbosity
+                -> FilePath    -- ^Where to look
                 -> IO FilePath -- <pkgname>.cabal
-findPackageDesc p = do ls <- getDirectoryContents p
-                       let descs = filter matchesDescFile ls
-                       descriptionCheck descs
+findPackageDesc verbosity p
+ = do ls <- getDirectoryContents p
+      let descs = filter matchesDescFile ls
+      descriptionCheck verbosity descs
 
 -- |Optional auxiliary package information file (/pkgname/@.buildinfo@)
 defaultHookedPackageDesc :: IO (Maybe FilePath)
@@ -455,12 +539,43 @@ findHookedPackageDesc
     -> IO (Maybe FilePath)	-- ^/dir/@\/@/pkgname/@.buildinfo@, if present
 findHookedPackageDesc dir = do
     ns <- getDirectoryContents dir
-    case [dir `joinFileName`  n |
-		n <- ns, snd (splitFileExt n) == buildInfoExt] of
+    case [dir </>  n |
+		n <- ns, takeExtension n == '.':buildInfoExt] of
 	[] -> return Nothing
 	[f] -> return (Just f)
 	_ -> die ("Multiple files with extension " ++ buildInfoExt)
 
+-- ------------------------------------------------------------
+-- * Platform file extensions
+-- ------------------------------------------------------------ 
+
+-- ToDo: This should be determined via autoconf (AC_EXEEXT)
+-- | Extension for executable files
+-- (typically @\"\"@ on Unix and @\"exe\"@ on Windows or OS\/2)
+exeExtension :: String
+exeExtension = case os of
+                   Windows _ -> "exe"
+                   _         -> ""
+
+-- ToDo: This should be determined via autoconf (AC_OBJEXT)
+-- | Extension for object files. For GHC and NHC the extension is @\"o\"@.
+-- Hugs uses either @\"o\"@ or @\"obj\"@ depending on the used C compiler.
+objExtension :: String
+objExtension = "o"
+
+-- | Extension for dynamically linked (or shared) libraries
+-- (typically @\"so\"@ on Unix and @\"dll\"@ on Windows)
+dllExtension :: String
+dllExtension = case os of
+                   Windows _ -> "dll"
+                   OSX       -> "dylib"
+                   _         -> "so"
+
+devNull :: FilePath
+devNull = case os of
+                   Windows _ -> "NUL"
+                   _         -> "/dev/null"
+
 -- ------------------------------------------------------------
 -- * Testing
 -- ------------------------------------------------------------
@@ -470,36 +585,19 @@ hunitTests :: [Test]
 hunitTests
     = let suffixes = ["hs", "lhs"]
           in [TestCase $
-#if mingw32_HOST_OS || mingw32_TARGET_OS
        do mp1 <- moduleToFilePath [""] "Distribution.Simple.Build" suffixes --exists
           mp2 <- moduleToFilePath [""] "Foo.Bar" suffixes    -- doesn't exist
           assertEqual "existing not found failed"
-                   ["Distribution\\Simple\\Build.hs"] mp1
+                   ["Distribution" </> "Simple" </> "Build.hs"] mp1
           assertEqual "not existing not nothing failed" [] mp2,
 
         "moduleToPossiblePaths 1" ~: "failed" ~:
-             ["Foo\\Bar\\Bang.hs","Foo\\Bar\\Bang.lhs"]
+             ["Foo" </> "Bar" </> "Bang.hs","Foo" </> "Bar" </> "Bang.lhs"]
                 ~=? (moduleToPossiblePaths "" "Foo.Bar.Bang" suffixes),
         "moduleToPossiblePaths2 " ~: "failed" ~:
               (moduleToPossiblePaths "" "Foo" suffixes) ~=? ["Foo.hs", "Foo.lhs"],
         TestCase (do files <- filesWithExtensions "." "cabal"
                      assertEqual "filesWithExtensions" "Cabal.cabal" (head files))
-#else
-       do mp1 <- moduleToFilePath [""] "Distribution.Simple.Build" suffixes --exists
-          mp2 <- moduleToFilePath [""] "Foo.Bar" suffixes    -- doesn't exist
-          assertEqual "existing not found failed"
-                   ["Distribution/Simple/Build.hs"] mp1
-          assertEqual "not existing not nothing failed" [] mp2,
-
-        "moduleToPossiblePaths 1" ~: "failed" ~:
-             ["Foo/Bar/Bang.hs","Foo/Bar/Bang.lhs"]
-                ~=? (moduleToPossiblePaths "" "Foo.Bar.Bang" suffixes),
-        "moduleToPossiblePaths2 " ~: "failed" ~:
-              (moduleToPossiblePaths "" "Foo" suffixes) ~=? ["Foo.hs", "Foo.lhs"],
-
-        TestCase (do files <- filesWithExtensions "." "cabal"
-                     assertEqual "filesWithExtensions" "Cabal.cabal" (head files))
-#endif
           ]
 
 -- |Might want to make this more generic some day, with regexps
@@ -513,5 +611,5 @@ filesWithExtensions dir extension
     = do allFiles <- getDirectoryContents dir
          return $ filter hasExt allFiles
     where
-      hasExt f = snd (splitFileExt f) == extension
+      hasExt f = takeExtension f == '.':extension
 #endif
