@@ -45,6 +45,7 @@ module Distribution.Version (
   -- * Package versions
   Version(..),
   showVersion,
+  readVersion,
   parseVersion,
 
   -- * Version ranges
@@ -54,6 +55,7 @@ module Distribution.Version (
   withinRange,
   showVersionRange,
   parseVersionRange,
+  isAnyVersion,
 
   -- * Dependencies
   Dependency(..),
@@ -68,11 +70,13 @@ import Data.Version	( Version(..), showVersion, parseVersion )
 #endif
 
 import Control.Monad    ( liftM )
+import Data.Char	( isSpace )
+import Data.Maybe	( listToMaybe )
 
 import Distribution.Compat.ReadP
 
 #ifdef DEBUG
-import HUnit
+import Test.HUnit
 #endif
 
 -- -----------------------------------------------------------------------------
@@ -192,6 +196,10 @@ parseVersion = do branch <- sepBy1 (liftM read $ munch1 isDigit) (char '.')
 
 #endif
 
+readVersion :: String -> Maybe Version
+readVersion str =
+  listToMaybe [ r | (r,s) <- readP_to_S parseVersion str, all isSpace s ]
+
 -- -----------------------------------------------------------------------------
 -- Version ranges
 
@@ -208,6 +216,10 @@ data VersionRange
   | IntersectVersionRanges  VersionRange VersionRange
   deriving (Show,Read,Eq)
 
+isAnyVersion :: VersionRange -> Bool
+isAnyVersion AnyVersion = True
+isAnyVersion _ = False
+
 orLaterVersion :: Version -> VersionRange
 orLaterVersion   v = UnionVersionRanges (ThisVersion v) (LaterVersion v)
 
