@@ -256,6 +256,7 @@ nmergeIO lss
 -- Bound Threads
 
 {- $boundthreads
+   #boundthreads#
 
 Support for multiple operating system threads and bound threads as described
 below is currently only available in the GHC runtime system if you use the
@@ -309,7 +310,11 @@ This means that you can use all kinds of foreign libraries from this thread
 
 Just to clarify, 'forkOS' is /only/ necessary if you need to associate
 a Haskell thread with a particular OS thread.  It is not necessary if
-you only need to make non-blocking foreign calls (see "Control.Concurrent#osthreads").
+you only need to make non-blocking foreign calls (see
+"Control.Concurrent#osthreads").  Neither is it necessary if you want
+to run threads in parallel on a multiprocessor: threads created with
+'forkIO' will be shared out amongst the running CPUs (using GHC,
+@-threaded@, and the @+RTS -N@ runtime option).
 
 -}
 forkOS :: IO () -> IO ThreadId
@@ -439,7 +444,7 @@ runInUnboundThread action = do
       worker OS threads.  If you need control over which particular OS
       thread is used to run a given Haskell thread, perhaps because
       you need to call a foreign library that uses OS-thread-local
-      state, then you need "bound threads" (see above).
+      state, then you need bound threads (see "Control.Concurrent#boundthreads").
 
       If you don't use the @-threaded@ option, then the runtime does
       not make use of multiple OS threads.  Foreign calls will block
@@ -491,7 +496,7 @@ runInUnboundThread action = do
 >    	    takeMVar m
 >    	    waitForChildren
 >    
->    forkChild :: IO () -> IO ()
+>    forkChild :: IO () -> IO ThreadId
 >    forkChild io = do
 >    	 mvar <- newEmptyMVar
 >    	 childs <- takeMVar children
