@@ -24,12 +24,13 @@ module Control.Monad.Fix (
 
 import Prelude
 import System.IO
+import Control.Monad.Instances ()
+import Data.Function (fix)
+#ifdef __HUGS__
+import Hugs.Prelude (MonadFix(mfix))
+#endif
 
--- | @'fix' f@ is the least fixed point of the function @f@,
--- i.e. the least defined @x@ such that @f x = x@.
-fix :: (a -> a) -> a
-fix f = let x = f x in x
-
+#ifndef __HUGS__
 -- | Monads having fixed points with a \'knot-tying\' semantics.
 -- Instances of 'MonadFix' should satisfy the following laws:
 --
@@ -54,6 +55,7 @@ class (Monad m) => MonadFix m where
 	-- output fed back as the input.  Hence @f@ should not be strict,
 	-- for then @'mfix' f@ would diverge.
 	mfix :: (a -> m a) -> m a
+#endif /* !__HUGS__ */
 
 -- Instances of MonadFix for Prelude monads
 
@@ -71,3 +73,6 @@ instance MonadFix [] where
 -- IO:
 instance MonadFix IO where
     mfix = fixIO 
+
+instance MonadFix ((->) r) where
+    mfix f = \ r -> let a = f a r in a
