@@ -89,7 +89,8 @@ struct Capability_ {
 #endif
 
     // Per-capability STM-related data
-    StgTVarWaitQueue *free_tvar_wait_queues;
+    StgTVarWatchQueue *free_tvar_watch_queues;
+    StgInvariantCheckQueue *free_invariant_check_queues;
     StgTRecChunk *free_trec_chunks;
     StgTRecHeader *free_trec_headers;
     nat transaction_tokens;
@@ -216,7 +217,7 @@ void prodAllCapabilities (void);
 // Waits for a capability to drain of runnable threads and workers,
 // and then acquires it.  Used at shutdown time.
 //
-void shutdownCapability (Capability *cap, Task *task);
+void shutdownCapability (Capability *cap, Task *task, rtsBool wait_foreign);
 
 // Attempt to gain control of a Capability if it is free.
 //
