@@ -1,13 +1,14 @@
 /* includes for compiling .cmm files via-C */
+#include "Stg.h"
 #include "Rts.h"
 #include "RtsFlags.h"
 #include "RtsUtils.h"
 #include "StgRun.h"
+#include "Storage.h"
 #include "Schedule.h"
 #include "Printer.h"
 #include "Sanity.h"
 #include "STM.h"
-#include "Storage.h"
 #include "SchedAPI.h"
 #include "Timer.h"
 #include "ProfHeap.h"
@@ -20,3 +21,4 @@
 #include "ThreadLabels.h"
 #include "Threads.h"
 #include "Prelude.h"
+#include "Stable.h"
