@@ -12,11 +12,11 @@
 #include "TSO.h"
 #include "Schedule.h"
 #include "RtsFlags.h"
-#include "Storage.h"
 #include "LdvProfile.h"
 #include "Updates.h"
 #include "Sanity.h"
 #include "Liveness.h"
+#include "Prelude.h"
 
 #include "Bytecodes.h"
 #include "Printer.h"
@@ -55,7 +55,6 @@
 
 #define BCO_PTR(n)    (W_)ptrs[n]
 #define BCO_LIT(n)    literals[n]
-#define BCO_ITBL(n)   itbls[n]
 
 #define LOAD_STACK_POINTERS					\
     Sp = cap->r.rCurrentTSO->sp;				\
@@ -85,6 +84,8 @@ allocate_NONUPD (int n_words)
     return allocate(stg_max(sizeofW(StgHeader)+MIN_PAYLOAD_SIZE, n_words));
 }
 
+int rts_stop_next_breakpoint = 0;
+int rts_stop_on_exception = 0;
 
 #ifdef INTERP_STATS
 
@@ -105,6 +106,7 @@ int it_ofreq[27];
 int it_oofreq[27][27];
 int it_lastopc;
 
+
 #define INTERP_TICK(n) (n)++
 
 void interp_startup ( void )
@@ -177,6 +179,9 @@ static StgWord app_ptrs_itbl[] = {
     (W_)&stg_ap_pppppp_info,
 };
 
+HsStablePtr rts_breakpoint_io_action; // points to the IO action which is executed on a breakpoint
+                                // it is set in main/GHC.hs:runStmt
+
 Capability *
 interpretBCO (Capability* cap)
 {
@@ -184,7 +189,7 @@ interpretBCO (Capability* cap)
     // that these entities are non-aliasable.
     register StgPtr       Sp;    // local state -- stack pointer
     register StgPtr       SpLim; // local state -- stack lim pointer
-    register StgClosure*  obj;
+    register StgClosure   *tagged_obj = 0, *obj;
     nat n, m;
 
     LOAD_STACK_POINTERS;
@@ -200,8 +205,8 @@ interpretBCO (Capability* cap)
     //      	+---------------+
     //       
     if (Sp[0] == (W_)&stg_enter_info) {
-	Sp++;
-	goto eval;
+       Sp++;
+       goto eval;
     }
 
     // ------------------------------------------------------------------------
@@ -219,7 +224,7 @@ interpretBCO (Capability* cap)
     //      	+---------------+
     //       
     else if (Sp[0] == (W_)&stg_apply_interp_info) {
-	obj = (StgClosure *)Sp[1];
+	obj = UNTAG_CLOSURE((StgClosure *)Sp[1]);
 	Sp += 2;
 	goto run_BCO_fun;
     }
@@ -236,9 +241,10 @@ interpretBCO (Capability* cap)
 
     // Evaluate the object on top of the stack.
 eval:
-    obj = (StgClosure*)Sp[0]; Sp++;
+    tagged_obj = (StgClosure*)Sp[0]; Sp++;
 
 eval_obj:
+    obj = UNTAG_CLOSURE(tagged_obj);
     INTERP_TICK(it_total_evals);
 
     IF_DEBUG(interpreter,
@@ -262,7 +268,7 @@ eval_obj:
     case IND_OLDGEN_PERM:
     case IND_STATIC:
     { 
-	obj = ((StgInd*)obj)->indirectee;
+	tagged_obj = ((StgInd*)obj)->indirectee;
 	goto eval_obj;
     }
     
@@ -286,8 +292,10 @@ eval_obj:
 	break;
 	
     case BCO:
+    {
 	ASSERT(((StgBCO *)obj)->arity > 0);
 	break;
+    }
 
     case AP:	/* Copied from stg_AP_entry. */
     {
@@ -300,7 +308,7 @@ eval_obj:
 	// Stack check
 	if (Sp - (words+sizeofW(StgUpdateFrame)) < SpLim) {
 	    Sp -= 2;
-	    Sp[1] = (W_)obj;
+	    Sp[1] = (W_)tagged_obj;
 	    Sp[0] = (W_)&stg_enter_info;
 	    RETURN_TO_SCHEDULER(ThreadInterpret, StackOverflow);
 	}
@@ -320,7 +328,7 @@ eval_obj:
 	    Sp[i] = (W_)ap->payload[i];
 	}
 
-	obj = (StgClosure*)ap->fun;
+	obj = UNTAG_CLOSURE((StgClosure*)ap->fun);
 	ASSERT(get_itbl(obj)->type == BCO);
 	goto run_BCO_fun;
     }
@@ -343,16 +351,17 @@ eval_obj:
 		 printObj(obj);
 	    );
 	Sp -= 2;
-	Sp[1] = (W_)obj;
+	Sp[1] = (W_)tagged_obj;
 	Sp[0] = (W_)&stg_enter_info;
 	RETURN_TO_SCHEDULER_NO_PAUSE(ThreadRunGHC, ThreadYielding);
     }
     }
 
     // ------------------------------------------------------------------------
-    // We now have an evaluated object (obj).  The next thing to
+    // We now have an evaluated object (tagged_obj).  The next thing to
     // do is return it to the stack frame on top of the stack.
 do_return:
+    obj = UNTAG_CLOSURE(tagged_obj);
     ASSERT(closure_HNF(obj));
 
     IF_DEBUG(interpreter,
@@ -413,8 +422,16 @@ do_return:
     case UPDATE_FRAME:
 	// Returning to an update frame: do the update, pop the update
 	// frame, and continue with the next stack frame.
+        //
+        // NB. we must update with the *tagged* pointer.  Some tags
+        // are not optional, and if we omit the tag bits when updating
+        // then bad things can happen (albeit very rarely).  See #1925.
+        // What happened was an indirection was created with an
+        // untagged pointer, and this untagged pointer was propagated
+        // to a PAP by the GC, violating the invariant that PAPs
+        // always contain a tagged pointer to the function.
 	INTERP_TICK(it_retto_UPDATE);
-	UPD_IND(((StgUpdateFrame *)Sp)->updatee, obj); 
+	UPD_IND(((StgUpdateFrame *)Sp)->updatee, tagged_obj); 
 	Sp += sizeofW(StgUpdateFrame);
 	goto do_return;
 
@@ -424,6 +441,8 @@ do_return:
 	INTERP_TICK(it_retto_BCO);
 	Sp--;
 	Sp[0] = (W_)obj;
+        // NB. return the untagged object; the bytecode expects it to
+        // be untagged.  XXX this doesn't seem right.
 	obj = (StgClosure*)Sp[2];
 	ASSERT(get_itbl(obj)->type == BCO);
 	goto run_BCO_return;
@@ -438,7 +457,7 @@ do_return:
 		 printStackChunk(Sp,cap->r.rCurrentTSO->stack+cap->r.rCurrentTSO->stack_size);
 	    );
 	Sp -= 2;
-	Sp[1] = (W_)obj;
+	Sp[1] = (W_)tagged_obj;
 	Sp[0] = (W_)&stg_enter_info;
 	RETURN_TO_SCHEDULER_NO_PAUSE(ThreadRunGHC, ThreadYielding);
     }
@@ -511,6 +530,7 @@ do_return_unboxed:
     // Application...
 
 do_apply:
+    ASSERT(obj == UNTAG_CLOSURE(tagged_obj));
     // we have a function to apply (obj), and n arguments taking up m
     // words on the stack.  The info table (stg_ap_pp_info or whatever)
     // is on top of the arguments on the stack.
@@ -524,7 +544,7 @@ do_apply:
 	    pap = (StgPAP *)obj;
 
 	    // we only cope with PAPs whose function is a BCO
-	    if (get_itbl(pap->fun)->type != BCO) {
+	    if (get_itbl(UNTAG_CLOSURE(pap->fun))->type != BCO) {
 		goto defer_apply_to_sched;
 	    }
 
@@ -549,7 +569,7 @@ do_apply:
 		for (i = 0; i < pap->n_args; i++) {
 		    Sp[i] = (W_)pap->payload[i];
 		}
-		obj = pap->fun;
+		obj = UNTAG_CLOSURE(pap->fun);
 		goto run_BCO_fun;
 	    } 
 	    else if (arity == n) {
@@ -557,7 +577,7 @@ do_apply:
 		for (i = 0; i < pap->n_args; i++) {
 		    Sp[i] = (W_)pap->payload[i];
 		}
-		obj = pap->fun;
+		obj = UNTAG_CLOSURE(pap->fun);
 		goto run_BCO_fun;
 	    } 
 	    else /* arity > n */ {
@@ -574,7 +594,7 @@ do_apply:
 		for (i = 0; i < m; i++) {
 		    new_pap->payload[pap->n_args + i] = (StgClosure *)Sp[i];
 		}
-		obj = (StgClosure *)new_pap;
+		tagged_obj = (StgClosure *)new_pap;
 		Sp += m;
 		goto do_return;
 	    }
@@ -616,7 +636,7 @@ do_apply:
 		for (i = 0; i < m; i++) {
 		    pap->payload[i] = (StgClosure *)Sp[i];
 		}
-		obj = (StgClosure *)pap;
+		tagged_obj = (StgClosure *)pap;
 		Sp += m;
 		goto do_return;
 	    }
@@ -626,7 +646,7 @@ do_apply:
 	default:
 	defer_apply_to_sched:
 	    Sp -= 2;
-	    Sp[1] = (W_)obj;
+	    Sp[1] = (W_)tagged_obj;
 	    Sp[0] = (W_)&stg_enter_info;
 	    RETURN_TO_SCHEDULER_NO_PAUSE(ThreadRunGHC, ThreadYielding);
     }
@@ -674,6 +694,7 @@ do_apply:
     // Sadly we have three different kinds of stack/heap/cswitch check
     // to do:
 
+
 run_BCO_return:
     // Heap check
     if (doYouWantToGC()) {
@@ -682,6 +703,7 @@ run_BCO_return:
     }
     // Stack checks aren't necessary at return points, the stack use
     // is aggregated into the enclosing function entry point.
+
     goto run_BCO;
     
 run_BCO_return_unboxed:
@@ -691,6 +713,7 @@ run_BCO_return_unboxed:
     }
     // Stack checks aren't necessary at return points, the stack use
     // is aggregated into the enclosing function entry point.
+
     goto run_BCO;
     
 run_BCO_fun:
@@ -717,6 +740,7 @@ run_BCO_fun:
 	Sp[0] = (W_)&stg_apply_interp_info; // placeholder, really
 	RETURN_TO_SCHEDULER(ThreadInterpret, StackOverflow);
     }
+
     goto run_BCO;
     
     // Now, actually interpret the BCO... (no returning to the
@@ -725,13 +749,11 @@ run_BCO:
     INTERP_TICK(it_BCO_entries);
     {
 	register int       bciPtr     = 1; /* instruction pointer */
-    register StgWord16 bci;
+        register StgWord16 bci;
 	register StgBCO*   bco        = (StgBCO*)obj;
 	register StgWord16* instrs    = (StgWord16*)(bco->instrs->payload);
 	register StgWord*  literals   = (StgWord*)(&bco->literals->payload[0]);
 	register StgPtr*   ptrs       = (StgPtr*)(&bco->ptrs->payload[0]);
-	register StgInfoTable** itbls = (StgInfoTable**)
-	    (&bco->itbls->payload[0]);
 
 #ifdef INTERP_STATS
 	it_lastopc = 0; /* no opcode */
@@ -757,6 +779,7 @@ run_BCO:
 		 //if (do_print_stack) checkStack(Sp,cap->r.rCurrentTSO->stack+cap->r.rCurrentTSO->stack_size,iSu);
 	    );
 
+
 	INTERP_TICK(it_insns);
 
 #ifdef INTERP_STATS
@@ -773,6 +796,103 @@ run_BCO:
 
     switch (bci & 0xFF) {
 
+        /* check for a breakpoint on the beginning of a let binding */
+        case bci_BRK_FUN: 
+        {
+            int arg1_brk_array, arg2_array_index, arg3_freeVars;
+            StgArrWords *breakPoints;
+            int returning_from_break;     // are we resuming execution from a breakpoint?
+                                          //  if yes, then don't break this time around
+            StgClosure *ioAction;         // the io action to run at a breakpoint
+
+            StgAP_STACK *new_aps;         // a closure to save the top stack frame on the heap
+            int i;
+            int size_words;
+
+            arg1_brk_array      = BCO_NEXT;  // 1st arg of break instruction
+            arg2_array_index    = BCO_NEXT;  // 2nd arg of break instruction
+            arg3_freeVars       = BCO_NEXT;  // 3rd arg of break instruction
+
+            // check if we are returning from a breakpoint - this info
+            // is stored in the flags field of the current TSO
+            returning_from_break = cap->r.rCurrentTSO->flags & TSO_STOPPED_ON_BREAKPOINT; 
+
+            // if we are returning from a break then skip this section
+            // and continue executing
+            if (!returning_from_break)
+            {
+               breakPoints = (StgArrWords *) BCO_PTR(arg1_brk_array);
+
+               // stop the current thread if either the
+               // "rts_stop_next_breakpoint" flag is true OR if the
+               // breakpoint flag for this particular expression is
+               // true
+               if (rts_stop_next_breakpoint == rtsTrue || 
+                   breakPoints->payload[arg2_array_index] == rtsTrue)
+               {
+                  // make sure we don't automatically stop at the
+                  // next breakpoint
+                  rts_stop_next_breakpoint = rtsFalse;
+
+                  // allocate memory for a new AP_STACK, enough to
+                  // store the top stack frame plus an
+                  // stg_apply_interp_info pointer and a pointer to
+                  // the BCO
+                  size_words = BCO_BITMAP_SIZE(obj) + 2;
+                  new_aps = (StgAP_STACK *) allocate (AP_STACK_sizeW(size_words));
+                  SET_HDR(new_aps,&stg_AP_STACK_info,CCS_SYSTEM); 
+                  new_aps->size = size_words;
+                  new_aps->fun = &stg_dummy_ret_closure; 
+
+                  // fill in the payload of the AP_STACK 
+                  new_aps->payload[0] = (StgClosure *)&stg_apply_interp_info;
+                  new_aps->payload[1] = (StgClosure *)obj;
+
+                  // copy the contents of the top stack frame into the AP_STACK
+                  for (i = 2; i < size_words; i++)
+                  {
+                     new_aps->payload[i] = (StgClosure *)Sp[i-2];
+                  }
+
+                  // prepare the stack so that we can call the
+                  // rts_breakpoint_io_action and ensure that the stack is
+                  // in a reasonable state for the GC and so that
+                  // execution of this BCO can continue when we resume
+                  ioAction = (StgClosure *) deRefStablePtr (rts_breakpoint_io_action);
+                  Sp -= 9;
+                  Sp[8] = (W_)obj;   
+                  Sp[7] = (W_)&stg_apply_interp_info;
+                  Sp[6] = (W_)&stg_noforceIO_info;     // see [unreg] below
+                  Sp[5] = (W_)new_aps;                 // the AP_STACK
+                  Sp[4] = (W_)BCO_PTR(arg3_freeVars);  // the info about local vars of the breakpoint
+                  Sp[3] = (W_)False_closure;            // True <=> a breakpoint
+                  Sp[2] = (W_)&stg_ap_pppv_info;
+                  Sp[1] = (W_)ioAction;                // apply the IO action to its two arguments above
+                  Sp[0] = (W_)&stg_enter_info;         // get ready to run the IO action
+                  // Note [unreg]: in unregisterised mode, the return
+                  // convention for IO is different.  The
+                  // stg_noForceIO_info stack frame is necessary to
+                  // account for this difference.
+
+                  // set the flag in the TSO to say that we are now
+                  // stopping at a breakpoint so that when we resume
+                  // we don't stop on the same breakpoint that we
+                  // already stopped at just now
+                  cap->r.rCurrentTSO->flags |= TSO_STOPPED_ON_BREAKPOINT;
+
+                  // stop this thread and return to the scheduler -
+                  // eventually we will come back and the IO action on
+                  // the top of the stack will be executed
+                  RETURN_TO_SCHEDULER_NO_PAUSE(ThreadRunGHC, ThreadYielding);
+               }
+            }
+            // record that this thread is not stopped at a breakpoint anymore
+            cap->r.rCurrentTSO->flags &= ~TSO_STOPPED_ON_BREAKPOINT;
+
+            // continue normal execution of the byte code instructions
+	    goto nextInsn;
+        }
+
 	case bci_STKCHECK: {
 	    // Explicit stack check at the beginning of a function
 	    // *only* (stack checks in case alternatives are
@@ -1019,12 +1139,12 @@ run_BCO:
 	    int i;
 	    int o_itbl         = BCO_NEXT;
 	    int n_words        = BCO_NEXT;
-	    StgInfoTable* itbl = INFO_PTR_TO_STRUCT(BCO_ITBL(o_itbl));
+	    StgInfoTable* itbl = INFO_PTR_TO_STRUCT(BCO_LIT(o_itbl));
 	    int request        = CONSTR_sizeW( itbl->layout.payload.ptrs, 
 					       itbl->layout.payload.nptrs );
 	    StgClosure* con = (StgClosure*)allocate_NONUPD(request);
 	    ASSERT( itbl->layout.payload.ptrs + itbl->layout.payload.nptrs > 0);
-	    SET_HDR(con, BCO_ITBL(o_itbl), CCS_SYSTEM/*ToDo*/);
+	    SET_HDR(con, (StgInfoTable*)BCO_LIT(o_itbl), CCS_SYSTEM/*ToDo*/);
 	    for (i = 0; i < n_words; i++) {
 		con->payload[i] = (StgClosure*)Sp[i];
 	    }
@@ -1145,7 +1265,7 @@ run_BCO:
 	    goto eval;
 
 	case bci_RETURN:
-	    obj = (StgClosure *)Sp[0];
+	    tagged_obj = (StgClosure *)Sp[0];
 	    Sp++;
 	    goto do_return;
 
@@ -1215,9 +1335,14 @@ run_BCO:
 	    // on the stack frame to describe this chunk of stack.
 	    //
 	    Sp -= ret_dyn_size;
-	    ((StgRetDyn *)Sp)->liveness = NO_PTRS | N_NONPTRS(stk_offset);
+	    ((StgRetDyn *)Sp)->liveness = R1_PTR | N_NONPTRS(stk_offset);
 	    ((StgRetDyn *)Sp)->info = (StgInfoTable *)&stg_gc_gen_info;
 
+            // save obj (pointer to the current BCO), since this
+            // might move during the call.  We use the R1 slot in the
+            // RET_DYN frame for this, hence R1_PTR above.
+            ((StgRetDyn *)Sp)->payload[0] = (StgClosure *)obj;
+
 	    SAVE_STACK_POINTERS;
 	    tok = suspendThread(&cap->r);
 
@@ -1238,6 +1363,16 @@ run_BCO:
 	    // And restart the thread again, popping the RET_DYN frame.
 	    cap = (Capability *)((void *)((unsigned char*)resumeThread(tok) - sizeof(StgFunTable)));
 	    LOAD_STACK_POINTERS;
+
+            // Re-load the pointer to the BCO from the RET_DYN frame,
+            // it might have moved during the call.  Also reload the
+            // pointers to the components of the BCO.
+            obj        = ((StgRetDyn *)Sp)->payload[0];
+            bco        = (StgBCO*)obj;
+            instrs     = (StgWord16*)(bco->instrs->payload);
+            literals   = (StgWord*)(&bco->literals->payload[0]);
+            ptrs       = (StgPtr*)(&bco->ptrs->payload[0]);
+
 	    Sp += ret_dyn_size;
 	    
 	    // Save the Haskell thread's current value of errno
@@ -1260,7 +1395,7 @@ run_BCO:
 	    bciPtr     = nextpc;
 	    goto nextInsn;
 	}
-
+ 
 	case bci_CASEFAIL:
 	    barf("interpretBCO: hit a CASEFAIL");
 	    
