@@ -87,13 +87,11 @@ register double fake_f9 __asm__("$f9");
 StgRegTable * StgRun(StgFunPtr f, StgRegTable *basereg STG_UNUSED)
 {
     while (f) {
-        /* XXX Disabled due to RtsFlags[]/RtsFlags mismatch
 	IF_DEBUG(interpreter,
 	    debugBelch("Jumping to ");
 	    printPtr((P_)f); fflush(stdout);
 	    debugBelch("\n");
 	    );
-        */
 	f = (StgFunPtr) (f)();
     }
     return (StgRegTable *)R1.p;
@@ -106,6 +104,12 @@ StgFunPtr StgReturn(void)
 
 #else /* !USE_MINIINTERPRETER */
 
+#ifdef LEADING_UNDERSCORE
+#define STG_RUN "_StgRun"
+#else
+#define STG_RUN "StgRun"
+#endif
+
 #ifdef LEADING_UNDERSCORE
 #define STG_RETURN "_StgReturn"
 #else
@@ -207,8 +211,8 @@ StgRunIsImplementedInAssembler(void)
 	/*
 	 * save callee-saves registers on behalf of the STG code.
 	 */
-	".globl StgRun\n"
-	"StgRun:\n\t"
+	".globl " STG_RUN "\n"
+	STG_RUN ":\n\t"
 	"subq %0, %%rsp\n\t"
 	"movq %%rsp, %%rax\n\t"
 	"addq %0-48, %%rax\n\t"
@@ -228,7 +232,7 @@ StgRunIsImplementedInAssembler(void)
         "movq %%rdi,%%rax\n\t"
         "jmp *%%rax\n\t"
 
-	".global " STG_RETURN "\n"
+	".globl " STG_RETURN "\n"
        	STG_RETURN ":\n\t"
 
 	"movq %%rbx, %%rax\n\t"   /* Return value in R1  */
