@@ -28,6 +28,15 @@ initThreadLabelTable(void)
   }
 }
 
+void
+freeThreadLabelTable(void)
+{
+    if (threadLabels != NULL) {
+        freeHashTable(threadLabels, NULL);
+        threadLabels = NULL;
+    }
+}
+
 void
 updateThreadLabel(StgWord key, void *data)
 {
