@@ -2,16 +2,16 @@
  *
  * (c) The GHC Team, 1995-2006
  *
- * Interval timer service for profiling and pre-emptive scheduling.
+ * Interface to the RTS timer signal (uses OS-dependent Ticker.h underneath)
  *
  * ---------------------------------------------------------------------------*/
 
 #ifndef TIMER_H
 #define TIMER_H
 
-typedef void (*TickProc)(int);
-
-extern int startTimer(void);
-extern int stopTimer(void);
+extern void initTimer(void);
+extern void startTimer(void);
+extern void stopTimer(void);
+extern void exitTimer(void);
 
 #endif /* TIMER_H */
