@@ -10,6 +10,7 @@
 /* #include "PosixSource.h" */
 
 #include "Rts.h"
+#include "Storage.h"
 #include "Schedule.h"
 #include "RtsUtils.h"
 #include "RtsFlags.h"
@@ -60,7 +61,7 @@ wakeUpSleepingThreads(lnat ticks)
     rtsBool flag = rtsFalse;
 
     while (sleeping_queue != END_TSO_QUEUE &&
-	   (int)(ticks - sleeping_queue->block_info.target) > 0) {
+	   (int)(ticks - sleeping_queue->block_info.target) >= 0) {
 	tso = sleeping_queue;
 	sleeping_queue = tso->link;
 	tso->why_blocked = NotBlocked;
@@ -140,11 +141,15 @@ awaitEvent(rtsBool wait)
       for(tso = blocked_queue_hd; tso != END_TSO_QUEUE; tso = next) {
 	next = tso->link;
 
+      /* On FreeBSD FD_SETSIZE is unsigned. Cast it to signed int
+       * in order to switch off the 'comparison between signed and
+       * unsigned error message
+       */
 	switch (tso->why_blocked) {
 	case BlockedOnRead:
 	  { 
 	    int fd = tso->block_info.fd;
-	    if (fd >= FD_SETSIZE) {
+	    if (fd >= (int)FD_SETSIZE) {
 		barf("awaitEvent: descriptor out of range");
 	    }
 	    maxfd = (fd > maxfd) ? fd : maxfd;
@@ -155,7 +160,7 @@ awaitEvent(rtsBool wait)
 	case BlockedOnWrite:
 	  { 
 	    int fd = tso->block_info.fd;
-	    if (fd >= FD_SETSIZE) {
+	    if (fd >= (int)FD_SETSIZE) {
 		barf("awaitEvent: descriptor out of range");
 	    }
 	    maxfd = (fd > maxfd) ? fd : maxfd;
