@@ -4,6 +4,7 @@
  *
  * (c) sof, 2002-2003.
  */
+
 #if !defined(THREADED_RTS)
 
 #include "Rts.h"
@@ -173,9 +174,9 @@ startupAsyncIO()
 }
 
 void
-shutdownAsyncIO()
+shutdownAsyncIO(rtsBool wait_threads)
 {
-    ShutdownIOManager();
+    ShutdownIOManager(wait_threads);
     if (completed_req_event != INVALID_HANDLE_VALUE) {
         CloseHandle(completed_req_event);
 	completed_req_event = INVALID_HANDLE_VALUE;
@@ -188,6 +189,7 @@ shutdownAsyncIO()
         CloseHandle(completed_table_sema);
 	completed_table_sema = NULL;
     }
+    DeleteCriticalSection(&queue_lock);
 }
 
 /*
@@ -356,4 +358,4 @@ resetAbandonRequestWait( void )
     ResetEvent(abandon_req_wait);
 }
 
-#endif
+#endif /* !defined(THREADED_RTS) */
